"""
A module that provides mixin classes to enable persistence with HDF5 files.

Author: Nils Geib, nils.geib@uni-jena.de
"""
import h5py
import lzma
import tempfile
import types
import os
from . import handlers
from . import options
try:
    import magic
except ModuleNotFoundError:
    magic = None

VERSION = 1.0


def save(val, path, archive=False, options=options.DEFAULT_OPTIONS):
    ''' Saves an object in an HDF5 file.

    Parameters
    ----------
    val : object
        Any Python value that is made up of storeable instances. Those are
        built-in types, numpy datatypes and types with custom handlers.
    path : str or Path instance
        Save path of the HDF5 file. Existing files will be overwritten!
    archive : bool, optional
        If ``True`` will compress the whole hdf5 file. This is useful when
        dealing with (many) small HDF5 files as those contain significant
        overhead.
    options : HDF5Options instance, optional
        The HDF5 options that will be used for saving. Defaults to the
        global options instance `DEFAULT_OPTIONS`.
    '''
    if archive:
        _savez(val, path, options=options)
    else:
        _save(val, path, options=options)


def _savez(val, path, options=options.DEFAULT_OPTIONS):
    ''' Saves an object in a compressed HDF5 file.
    '''
    # This is not a secure way to write and read from a tempfile.
    # But due to h5py's limitations I did not find a better way.
    tmppath = tempfile.mktemp()
    _save(val, tmppath, options)
    with open(tmppath, 'rb') as f:
        with lzma.LZMAFile(path, 'w') as zf:
            zf.write(f.read())
    os.remove(tmppath)


def _save(val, path, options=options.DEFAULT_OPTIONS):
    ''' Saves an object in an HDF5 file.

    Parameters
    ----------
    val : object
        Any Python value that is made up of storeable instances. Those are
        built-in types, numpy datatypes and types with custom handlers.
    path : str or Path instance
        Save path of the HDF5 file. Existing files will be overwritten!
    options : HDF5Options instance
        The HDF5 options that will be used for saving. Defaults to the
        global options instance `DEFAULT_OPTIONS`.
    '''
    with h5py.File(path, 'w', libver=options.libver, driver=options.driver,
                   **options.kwds) as f:
        f.attrs['__version__'] = VERSION
        handlers.save_to_level(val, f, options, name=None)


def _load(path, obj=None):
    ret = None
    with h5py.File(path, 'r') as f:
        if '__version__' not in f.attrs:
            raise ValueError('File `%s` is not a valid serialized file.' %
                             path)
        if f.attrs['__version__'] != VERSION:
            raise ValueError('File `%s` was created with a different version.'
                             % path)
        # special case: A single dataset stored at the top-level file group.
        # In this case there is no '__class__'-attribute and we call the
        # load function directly on this single dataset.
        keys = list(f.keys())
        if len(keys) == 1:
            ds = f[keys[0]]
            if isinstance(ds, h5py.Dataset):
                ret = handlers.load_from_level(ds, obj=obj)
        # General case: recursively load the object structure from the file
        # starting with the root group.
        if ret is None:
            ret = handlers.load_from_level(f, obj=obj)
    return ret


def _loadz(path, obj=None):
    tmppath = tempfile.mktemp()
    with open(tmppath, 'wb') as f:
        with lzma.LZMAFile(path, 'r') as zf:
            f.write(zf.read())
    ret = load(tmppath, obj=obj)
    os.remove(tmppath)
    return ret


def load(path, obj=None, archive=None):
    ''' Reads a possibly compressed HDF5 file.

    If archive is ``None`` it is retrieved with python-magic.
    '''
    if archive is None:
        if magic is None:
            archive = False  # simply assume non-compressed file - may fail
        else:
            mime = magic.from_file(str(path))
            archive = (mime == 'XZ compressed data')
    if archive:
        return _loadz(path, obj=obj)
    return _load(path, obj=obj)


class IOHandler(handlers.InstanceHandler):
    level_type = 'group'
    instances = []  # instances are filled by MetaIO

    def save(cls, val, level, options, name):
        attrs = val.to_dict()
        for key, value in attrs.items():
            handlers.save_to_level(value, level, options, key)

    def load_from_level(cls, level, obj=None):
        attrs = handlers.DictHandler.load_from_level(level)
        if obj is None:
            # call from_dict on class
            t = cls.get_type(level)
            return t.from_dict(attrs)
        # else call update_from_dict on object
        obj.update_from_dict(attrs)


class MetaIO(type):
    ''' All sub-classes of IO are automatically registered with the default
        IOHandler. This allows easy storage by simply subclassing IO.
    '''
    def __new__(cls, clsname, bases, attrs):
        newclass = super().__new__(cls, clsname, bases, attrs)
        IOHandler.register(newclass)
        return newclass


class IO(metaclass=MetaIO):
    """ Provides an interface for saving to and loading from a HDF5 file.

    This class can be mixed-in to easily add persistence to your existing
    Python classes. By default all attributes of an object will be stored.
    Upon loading these attributes will be loaded and `__init__` will *not*
    be called.

    Often a better way is to store only the necessary attributes by
    giving a list of attribute names in the private attribute `_io_store`.
    Then you have to overwrite the `_post_init()` method that initializes
    your object from these stored attributes. It is usually also be called at
    the end of the original `__init__` and should not mean extra effort.

    Lastly, you can simply overwrite `load_from_dict` to implement a
    completely custom loader.
    """
    # A list of attribute names that are not stored
    _io_store_not = []
    # None or a list of attribute names that are stored exclusively.
    _io_store = None

    def __new__(cls, *args, **kwargs):
        self = super(IO, cls).__new__(cls)
        return self

    def save(self, path, archive=False, options=options.DEFAULT_OPTIONS):
        save(self, path, archive=archive, options=options)

    def save_to_group(self, g, name):
        handlers.save_to_level(self, g, self._io_options, name)

    def to_dict(self):
        # local rename
        io_store = self._io_store
        io_store_not = self._io_store_not
        cls = self.__class__
        # make sure that _io_store and _io_store are stored if they differ from
        # the class defaults
        if io_store is not None:
            if (io_store != cls._io_store and
                    '_io_store' not in io_store):
                io_store = io_store + ['_io_store']  # make copy
            if (io_store_not != cls._io_store_not and
                    '_io_store_not' not in io_store):
                io_store = io_store + ['_io_store_not']  # make copy
        # actually store the attributes
        attrs = dict()
        for key, value in self.__dict__.items():
            if io_store is not None:
                if key not in io_store:
                    continue
            elif (key in io_store_not or key[:2] == '__' or
                    isinstance(value, types.MethodType)):
                continue
            attrs[key] = value
        return attrs

    @classmethod
    def load(cls, path):
        ret = load(path)
        if not isinstance(ret, cls):
            raise ValueError('File does not store instance of class '
                             '`%s` but of `%s`.' % (str(type(ret)), str(cls)))
        return ret

    @classmethod
    def load_from_group(cls, group):
        ret = handlers.load_from_level(group)
        if not isinstance(ret, cls):
            raise ValueError('Group does not store instance of class '
                             '`%s` but of `%s`.' % (str(type(ret)), str(cls)))
        return ret

    @classmethod
    def from_dict(cls, attrs):
        obj = cls.__new__(cls)
        obj.update_from_dict(attrs)
        return obj

    def update_from_dict(self, attrs):
        for key, value in attrs.items():
            setattr(self, key, value)
        self._post_init()

    def _post_init(self):
        """ Hook to initialize an object from storage.
        """
        pass

    def _add_to_storage(self, key):
        if self._io_store is None:
            self._io_store = []
        self._io_store.append(key)

    def update_from_group(self, group):
        handlers.load_from_level(group, obj=self)

    def update(self, path):
        load(path, obj=self)





#%% Import modules
import femtoQ.tools as fq
import numpy as np
import matplotlib.pyplot as plt
from femtoQ.pr_backend import tdsilib as library_2dsi
from femtoQ.pr_backend import froglib as library_frog
from scipy.constants import c as C
from scipy.interpolate import interp1d as interp
from scipy.integrate import simps as simpson
import pypret

def freq2time(frequency, amplitude_spectrum):
    # Interpolate over new frequency grid, including negative w components
    freq_max = frequency[-1]
    freq_min = -freq_max
    N = len(amplitude_spectrum)
    dv = freq_max/N
    
    new_freq_grid = np.hstack((np.array([0]),np.linspace(dv,freq_max,N),np.linspace(freq_min,-dv,N)))
    
    spectrum_interpolated = interp(frequency,amplitude_spectrum,'quadratic',bounds_error=False,fill_value=0)(new_freq_grid)
     

    amplitude_time = np.fft.fftshift(np.fft.ifft(spectrum_interpolated))
    t = np.fft.fftshift(np.fft.fftfreq(2*N+1,dv))

    return t, amplitude_time


def twodsi(filename,  upconvWavelength = 'auto', wavelengthCutoffs = None, smoothSpectrum = False, zeroPadTrace = True, windowTrace = True, minTimeResolution = 1e-15, folder = '', spectrum = None):

    #%% Analysis parameters
    debug = False                # Setting True will output many more figures for debugging
    simulatedData = False        # Setting to True will load last set of simulated data rather than experimental data
    fitPolynomialToPhase = False # Setting to an integer value >1 will fit a polynomial of same order to the phase curve in w space
    simulateFrogTrace = False    # Simulate a Frog trace with final result
    
    
    
    
    """ If file contains data acquired over a long distance on the stage, with a shear
        that changes a lot from start to end, set a value here to window the trace
        around the point where the shear is equal to that value """
    manualShear = None
    
    
    """ Range of stage displacement over which to window the trace. 
        Set to None to keep full data. """
    zWindowLength = None
    
    data = np.load(folder+filename)
    fixedMirrorData = data['upconvSpectrum']
    wavelengths = data['wavelengths']
    shearStagePosition = data['shearStagePosition']
    movingMirrorData = data['shearTrace']
    twoDSIStagePosition = data['twoDSIStagePosition']
    tdsiTrace = data['twoDSITrace']
    
    #%% Load data
    if simulatedData:
        tdsiTrace = data['twoDSITrace']
        timeVector = data['timeVector']
        inputPulse = data['inputPulse']
    
    data.close()
    
    
    #%% Cut spectrum to relevant spectral region
    """ Min and max wavelengths of spectral windowing. Any data outisde of this range
        will be discarded. """
        
    if wavelengthCutoffs is not None:
        minWavelength = wavelengthCutoffs[0]
        maxWavelength = wavelengthCutoffs[1]
        
        fixedMirrorData, movingMirrorData, wavelengths, tdsiTrace = library_2dsi.cut_spectrum(minWavelength, maxWavelength, fixedMirrorData, movingMirrorData, wavelengths, tdsiTrace)
    
    
    #%% Smooth spectrum, if desired
    if smoothSpectrum:
        fixedMirrorData = fq.ezsmooth(fixedMirrorData, window_len = 15, window = 'hanning')
        for ii in range(movingMirrorData.shape[0]):
            movingMirrorData[ii,:] = fq.ezsmooth(movingMirrorData[ii,:], window_len = 15, window = 'hanning')
    
    
    #%% Substract noise floor from data. Taken as max value of either the first or
    #   the last datapoint of the spectrum. Any negative value is afterward set to 0.
    fixedMirrorData, movingMirrorData = library_2dsi.substractNoiseFloor(fixedMirrorData, movingMirrorData)
    
    """ Add trace postprocessing here. Mostly consists of interpolating trace to a linear
        grid along stage displacement axis. Will be added once the 2dsi control software
        will output the actual stage positions. For now, stage linearity is assumed. """
    
    zMin = shearStagePosition[0]
    zMax = shearStagePosition[-1]
    
    movingMirrorZ = np.linspace(zMin,zMax,len(shearStagePosition))
    
    for ii in range(movingMirrorData.shape[1]):
        
        movingMirrorData[:,ii] = np.interp(movingMirrorZ, shearStagePosition, movingMirrorData[:,ii])
    
    zMin = twoDSIStagePosition[0]
    zMax = twoDSIStagePosition[-1]
    
    tdsiTraceZ = np.linspace(zMin,zMax,len(twoDSIStagePosition))
    
    for ii in range(movingMirrorData.shape[1]):
        
        tdsiTrace[:,ii] = np.interp(tdsiTraceZ, twoDSIStagePosition, tdsiTrace[:,ii])
    
    
    
    
    #%% Extract vector data form matrices
    wavelengths = wavelengths * 1e-9
    upconvPowerSpectrum = fixedMirrorData
    
    
    if spectrum is not None:
        data2 = np.load(folder+spectrum)
        fundPowerSpectrum = data2['spectrum']
        fundWavelengths = data2['wavelengths']*1e-9
        data2.close()
        if upconvWavelength == 'auto':
            upconvWavelength = library_2dsi.find_upconversion(wavelengths, upconvPowerSpectrum, fundWavelengths, fundPowerSpectrum)
            print('Upconversion wavelength: ' + str(round(upconvWavelength*1e9)) + ' nm')
        plt.figure()
        plt.plot(fundWavelengths*1e9,fundPowerSpectrum/fundPowerSpectrum.max(), label = 'Fundamental spectrum')
        plt.plot(1/(1/wavelengths-1/upconvWavelength)*1e9, upconvPowerSpectrum/upconvPowerSpectrum.max(),'--r', label = 'Un-upconverted spectrum')
        plt.ylim([0,1.05])
        plt.xlabel('Wavelengths [nm]')
        plt.ylabel('Normalized power density')
        plt.legend()
        
        
    if upconvWavelength == 'auto':
    
        upconvWavelength = wavelengths[np.argmax(upconvPowerSpectrum)]*2
        print('Upconversion wavelength: ' + str(round(upconvWavelength*1e9)) + ' nm')
        
        
    #%% Smooth spectrum, if desired
    # =============================================================================
    # if smoothingSpectrum:
    #     upconvPowerSpectrum = fq.ezsmooth(upconvPowerSpectrum, window_len = 7, window = 'flat')
    #     for ii in range(movingMirrorData.shape[0]):
    #         movingMirrorData[ii,:] = fq.ezsmooth(movingMirrorData[ii,:], window_len = 7, window = 'flat')
    # 
    # =============================================================================
    #%% Define linear space of stage displacement
    
    #%% Set trace windowing parameters
    if manualShear is not None:
    
        # Make a map of shear vs stage position
        shearMap = library_2dsi.make_shear_map(wavelengths, upconvPowerSpectrum, movingMirrorData,movingMirrorZ,debug)
        
        # Make a linear fit
        p = np.polyfit(shearMap, movingMirrorZ,1)
        
        # Find position of desired shear
        z0 = np.polyval(p, manualShear)
        
        # Set center of window  to center of data if desired shear can't be obtained
        if (z0 > zMax) or (z0 < zMin):
            z0 = (zMax + zMin)/2
    else:
        z0 = (zMin+zMax)/2
    
    
    #%% Window the trace, if desired
    if zWindowLength is not None:
        zMin = z0 - zWindowLength/2
        zMax = z0 + zWindowLength/2
        tdsiTrace = tdsiTrace[((tdsiTraceZ>=zMin) & (tdsiTraceZ<=zMax)),:]
        tdsiTraceZ = tdsiTraceZ[((tdsiTraceZ>=zMin) & (tdsiTraceZ<=zMax))]
        movingMirrorData = movingMirrorData[((movingMirrorZ>=zMin) & (movingMirrorZ<=zMax)),:]
        movingMirrorZ = movingMirrorZ[((movingMirrorZ>=zMin) & (movingMirrorZ<=zMax))]
    
    
    """ Find mean shear of data """ 
    shearFrequency = library_2dsi.find_shear(wavelengths, upconvPowerSpectrum, movingMirrorData,movingMirrorZ,tdsiTraceZ,debug)
    
    """ Perform 1D Fourier transform of trace along stage position dimension """ 
    FFTspatialFreq, FFTamplitude, FFTphase = library_2dsi.make1Dfft(wavelengths,tdsiTraceZ,tdsiTrace,zeropadding = zeroPadTrace, windowing = windowTrace, debugGraphs=debug)
    
    """ Cut FFT data to region with minimum amplitude """ 
    FFTamplitude, FFTphase, FFTwavelengths = library_2dsi.cut_fft(FFTamplitude, FFTphase, wavelengths,debug)
    
    if spectrum is not None:
        
        FFTwavelengths = 1 / ( 1/FFTwavelengths - 1/upconvWavelength )
        wavelengths = fundWavelengths
        upconvPowerSpectrum = fundPowerSpectrum
        upconvWavelength = np.infty
    
    """ Calculate spectral phase from FFT. Currently only using concatenation algorithm.
        Also outputs GDD and TOD (this calculation wil be moved to a new function 
        once other algorithms will be implemented """ 
    concW, concPhase,concGD, concGDD, concTOD, midpointW, midpointPhase,midpointGD, midpointGDD, midpointTOD = library_2dsi.calc_spectral_phase(shearFrequency, upconvWavelength,FFTamplitude, FFTphase, FFTwavelengths,polyfit = fitPolynomialToPhase,debugGraphs = debug)
    
    """ Calculate temporal enveloppe of pulse from upconverted power spectrum and
        spectral phase calculated above. Function directly outputs the square of 
        the enveloppe, to obtain intensity-like vertical units """
    tConc, pulseConc,Econc = library_2dsi.calc_temporal_envelope(wavelengths, upconvPowerSpectrum, upconvWavelength, concW, concPhase, True, minTimeResolution)
    tMidpoint, pulseMidpoint, Emidpoint = library_2dsi.calc_temporal_envelope(wavelengths, upconvPowerSpectrum, upconvWavelength, midpointW, midpointPhase, True,minTimeResolution)
    
    """ Calculate temporal enveloppe of Fourier-limited pulse """
    tLim, pulseLim, Elim = library_2dsi.calc_temporal_envelope(wavelengths, upconvPowerSpectrum, upconvWavelength, concW, np.zeros_like(concW), False,minTimeResolution)
    
    """ Realign pulse's peak intensity to t = 0 fs """
    tpeak = tConc[ np.argmax(pulseConc)]
    pulseConc = np.interp(tConc, tConc-tpeak, pulseConc)
    Econc = np.interp(tConc, tConc-tpeak, Econc)
    
    tpeak = tMidpoint[ np.argmax(pulseMidpoint)]
    pulseMidpoint = np.interp(tMidpoint, tMidpoint-tpeak, pulseMidpoint)
    
    tpeak = tLim[ np.argmax(pulseLim)]
    pulseLim = np.interp(tLim, tLim-tpeak, pulseLim)
    
    if simulatedData:
        tpeak = timeVector[ np.argmax(inputPulse)]
        inputPulse = np.interp(timeVector, timeVector-tpeak, inputPulse)
    
    """ Make figures and output values in console. Will be cleaned up in next update """
    fig = plt.figure()
    ax = fig.gca()
    ax.plot(tConc*1e15, pulseConc/np.max(pulseConc), '--b', label = 'Concat.')
    ax.plot(tMidpoint*1e15, pulseMidpoint/np.max(pulseMidpoint), '--r', label = 'Midpoint')
    ax.plot(tLim*1e15, pulseLim/np.max(pulseLim), 'k', label = 'Fourier lim.')
    ax.set_ylim([0,1.05])
    ax.set_xlabel('Time [fs]')
    ax.set_ylabel('Normalised intensity')
    if simulatedData:
        ax.plot(timeVector*1e15, inputPulse, '-.g', label = 'Sim. input')
        ax.set_ylim([0,1.05])
    ax.legend()
    ax.set_xlim([-150, 150])
    
# =============================================================================
#     fig = plt.figure()
#     ax = fig.gca()
#     ax.plot(tLim*1e15, pulseLim/np.max(pulseLim), 'k', label = 'Fourier limited')
#     ax.plot(tConc*1e15, pulseConc/np.max(pulseConc), '--r', label = 'Reconstructed')
#     ax.set_xlabel('Time [fs]')
#     ax.set_ylabel('Normalised intensity')
#     ax.legend()
#     ax.set_xlim([-150, 150])
# =============================================================================
    
    
    fig = plt.figure()
    axLeft = fig.gca()
    IIphase = np.argsort(concW,axis=None)
    IIGDD = np.argsort(concW[:,1:-1],axis=None)
    IITOD = np.argsort(concW[:,2:-2],axis=None)
    lambdaPhase = np.flip( 2*np.pi*2.998e8/concW.flatten()[IIphase]*1e9 )
    lambdaGDD = np.flip( 2*np.pi*2.998e8/concW[:,1:-1].flatten()[IIGDD]*1e9 )
    lambdaTOD = np.flip( 2*np.pi*2.998e8/concW[:,2:-2].flatten()[IITOD]*1e9 )
    concPhase =  np.flip( concPhase.flatten()[IIphase] )
    concGD =  np.flip( concGD.flatten()[IIGDD]*1e15 )
    concGDD =  np.flip( concGDD.flatten()[IIGDD]*1e30 )
    concTOD =  np.flip( concTOD.flatten()[IITOD]*1e45 )
    
    IIplot = upconvPowerSpectrum > np.max(upconvPowerSpectrum)/20
    wav1 = (1 / (1/wavelengths - 1/upconvWavelength)*1e9)[IIplot][0]
    wav2 = (1 / (1/wavelengths - 1/upconvWavelength)*1e9)[IIplot][-1]
    IIGDD = ((lambdaGDD>=wav1) & (lambdaGDD<=wav2))
    IIphase = ((lambdaPhase>=wav1) & (lambdaPhase<=wav2))
    w1 = 2*np.pi*2.998e8/wav2*1e9
    w2 = 2*np.pi*2.998e8/wav1*1e9
    
    
    axLeft.plot(1 / (1/wavelengths - 1/upconvWavelength)*1e9, upconvPowerSpectrum/np.max(upconvPowerSpectrum), '-b', linewidth = 2)
    axLeft.set_ylim([0,1.05])
    axRight = axLeft.twinx()
    axRight.plot(lambdaGDD[IIGDD], concGD[IIGDD], '--r', linewidth = 2, label = 'Concat.')
    axRight.plot(2*np.pi*2.998e8/midpointW[1:-1][((midpointW[1:-1]>w1) & (midpointW[1:-1]<w2))]*1e9, midpointGD[((midpointW[1:-1]>w1) & (midpointW[1:-1]<w2))]*1e15, '--k', linewidth = 2, label = 'Midpoint')
    axLeft.set_xlabel('Wavelength [nm]')
    axLeft.set_ylabel('Normalized intensity')
    axRight.set_ylabel(r'Group delay [fs]')
    axRight.legend()
    
    fig2 = plt.figure()
    axLeft2 = fig2.gca()
    axLeft2.plot(1 / (1/wavelengths - 1/upconvWavelength)*1e9, upconvPowerSpectrum / np.max(upconvPowerSpectrum), '-b', linewidth = 2)
    axLeft2.set_ylim([0,1.05])
    axRight2 = axLeft2.twinx()
    axRight2.plot(lambdaPhase[IIphase], concPhase[IIphase], '--r', linewidth = 2)
    axLeft2.set_xlabel('Wavelength [nm]')
    axLeft2.set_ylabel('Normalized power spectrum')
    axRight2.set_ylabel(r'Spectral phase [rad]')
    axLeft2.yaxis.label.set_color('blue')
    axRight2.yaxis.label.set_color('red')
    axRight2.set_ylim(-np.pi,np.pi)
    
    weightsGDD = np.interp( lambdaGDD, 1 / (1/wavelengths - 1/upconvWavelength)*1e9, upconvPowerSpectrum )
    averageGDD = np.average(concGDD, weights =  weightsGDD)
    weightsTOD = np.interp( lambdaTOD, 1 / (1/wavelengths - 1/upconvWavelength)*1e9, upconvPowerSpectrum )
    averageTOD = np.average(concTOD, weights =  weightsTOD)
    
    
    
    
    
    if  np.isnan(fq.ezfindwidth(tConc, pulseConc)):
        print( "Conc. pulse duration: NaN"  )
    else:
        print( "Conc. pulse duration: " + str(round(fq.ezfindwidth(tConc, pulseConc)*1e15*100)/100) +" fs" )
    
    if  np.isnan(fq.ezfindwidth(tMidpoint, pulseMidpoint)):
        print( "Midpoint pulse duration: NaN"  )
    else:
        print( "Midpoint pulse duration: " + str(round(fq.ezfindwidth(tMidpoint, pulseMidpoint)*1e15*100)/100) +" fs" )
        
    print( "F. lim. pulse duration: " + str(round(fq.ezfindwidth(tLim, pulseLim)*1e15*100)/100) +" fs")
    print('Average GDD (conc.): ' + str(round(averageGDD*100)/100) + ' fs sq.') 
    print('Average TOD (conc.): ' + str(round(averageTOD*100)/100) + ' fs cu.') 
    
    print('Shear frequency:' + str(round(shearFrequency/1e12 * 100)/100) + ' THz')
    
    if simulateFrogTrace:
        library_2dsi.simFrogTrace(tConc,Econc,minWavelength,maxWavelength)
    
    plt.show()
    
    return lambdaPhase, concPhase, tConc, Econc
    
    
def shgFROG(filename, initialGuess = 'gaussian', tau = None, method = 'copra', dt = None , maxIter = 100, symmetrizeGrid = False, wavelengthLimits = [0,np.inf], gridSize = None, marginalCorrection = None):
    
    
    delays, wavelengths, trace = library_frog.unpack_data(filename,wavelengthLimits)

    """ Recenter the trace to zero delay. Otherwise copra behaves weirdly"""
    marginal_t = simpson(trace,wavelengths,axis = 1)
    t_0 = delays[np.argmax(marginal_t)]
    delays -= t_0
    
    """ Removing negative values from trace. Seems to give slightly better results"""
    trace[trace<0] = 0
    
    """ PCGPA algorithm requires a symmetric grid """
    if method.lower() == 'pcgpa':
        symmetrizeGrid = True
    
    """ Adjust grid size if required """
    if symmetrizeGrid:
        if gridSize is None:
            gridSize = [len(wavelengths),len(wavelengths)]
        else:
            gridSize[1] = gridSize[0]
        
    
    
    if dt is None:
        dt = np.mean(np.diff(delays))
        
    if gridSize is not None:
        
        symTrace = np.zeros((gridSize[0],len(wavelengths)))
        
        delayLim = np.min([abs(delays[0]),abs(delays[-1])])
        
        # Symmetrize delay grid wrt frequency grid
        symDelays = np.linspace(-delayLim,delayLim,gridSize[0])
        
        for ii, wavelength in enumerate(wavelengths):
            interpTrace = interp(delays,trace[:,ii],'quadratic',bounds_error=False,fill_value=0)
            symTrace[:,ii] = interpTrace(symDelays)
            symTrace[:,ii] = 0.5*interpTrace(np.hstack((symDelays[symDelays<=0],symDelays[symDelays<0][-1::-1]))) + 0.5*interpTrace(np.hstack((symDelays[symDelays>=0][-1::-1],symDelays[symDelays>0])))
            
        trace = symTrace
        delays = symDelays
        dt = np.mean(np.diff(delays))
    
    """ Define time/frequency grids for input pulse """
    if gridSize is not None:
        ft = pypret.FourierTransform(gridSize[1], dt = dt)
    else:
        ft = pypret.FourierTransform(len(delays), dt = dt)
    
    """ Integrate over delay axis"""
    marginal_w = simpson(trace,delays,axis = 0)
    
        
    """ Carrier wavelength """
    lambda_0 = C/(simpson(C/wavelengths[-1::-1]*marginal_w[-1::-1],C/wavelengths[-1::-1],axis = 0)/simpson(marginal_w[-1::-1],C/wavelengths[-1::-1],axis = 0)) * 2
    
    """ Marginal correction: compare frequnecy marginal to spectrum autoconvolution.
        Relative differences between the two should correspond to experimental
        bandwidth limitation. Trace is adjusted accordingly to offset this effect. """
    if marginalCorrection is not None:
        data = np.load(marginalCorrection)
        corrWavelengths = data['wavelengths']*1e-9
        corrSpectrum = data['spectrum']
        
        corrSpectrumRaw = np.copy(corrSpectrum)
        
        corrW = np.linspace(-4*np.pi*C/corrWavelengths[0],4*np.pi*C/corrWavelengths[0],4*len(corrWavelengths)+1)
        
        corrSpectrum = interp( 2*np.pi*C/corrWavelengths[-1::-1], corrSpectrum[-1::-1] ,bounds_error=False,fill_value=0)(corrW)
        
        x,y = fq.ezifft(corrW,corrSpectrum)
        absc_conv,autoConv = fq.ezfft(x,y**2,neg = True) 
        autoConv = np.real(autoConv)

        
        autoConv = interp(absc_conv, autoConv,bounds_error=False,fill_value=0)(2*np.pi*C/wavelengths)
        
        marginal_w_corr = np.copy(marginal_w)
        marginal_w_corr[marginal_w_corr<=0] = marginal_w_corr[marginal_w_corr>0].min()
        marginal_w_corr = fq.ezsmooth(marginal_w_corr,15,'hanning')
        marginalCorr = ( autoConv/autoConv.max() ) / ( marginal_w_corr / marginal_w_corr.max() )

        for ii, delay in enumerate(delays):
            trace[ii,:]*=marginalCorr
        
        plt.figure()
        plt.plot(wavelengths*1e9,autoConv/autoConv.max(),label = 'From spectrum')
        plt.plot(wavelengths*1e9,marginal_w_corr / marginal_w_corr.max(), label = 'From FROG trace')
        plt.xlabel('Wavelengths [nm]')
        plt.ylabel('Frequency margianal')
        plt.ylim([0,1.05])
        plt.legend()
        
        plt.figure()
        plt.plot(wavelengths*1e9,marginalCorr)
        plt.xlabel('Wavelengths [nm]')
        plt.ylabel('Marginal correction factor')
        plt.ylim(bottom=0)
    
    """ Instantiate a pulse object w/ appropriate carrier wavelength
        (other parameters don't matter here)"""
    pulseP = pypret.Pulse(ft, lambda_0)
    pypret.random_gaussian(pulseP, 1e-15, phase_max=0.0)
        
    """ Instantiate a PNPS object for SHG-FROG technique"""
    pnps = pypret.PNPS(pulseP, "frog", "shg")
    pnps.calculate(pulseP.spectrum, delays)
    
    """ Export SHG frequency grid """
    w_shg = pnps.process_w
    w_fund = pulseP.w+pulseP.w0
    
    
    
    
    """ Interpolate trace to shg frequency grid """
    trace_w = np.zeros((len(delays),len(w_shg)))
    for ii,delay in enumerate(delays):  
    
        interpTrace = interp(C/wavelengths[-1::-1],trace[ii,:][-1::-1],'quadratic',bounds_error=False,fill_value=0)
        trace_w[ii,:] = interpTrace(w_shg/2/np.pi)
        
        
    
    
    """ Plot interpolated trace (to check interpolation errors) """
    plt.figure()
    plt.pcolormesh(delays*1e15,(2*np.pi*C/w_shg)*1e9,trace_w.transpose())
    if marginalCorrection is None:
        plt.title('Input trace (interpolated)')
    else:
        plt.title('Input trace (corrected + interpolated)')
    plt.xlabel('Delay [fs]')
    plt.ylabel('Wavelengths [nm]')
    plt.ylim(wavelengths[0]*1e9,wavelengths[-1]*1e9)
    plt.colorbar()

    """ Reformat trace for retriever """
    traceInput = pypret.mesh_data.MeshData(trace_w,delays,w_shg,labels = ['Delay','Frequency',''])


    """ Initial guess for iterative algorithm. Three options:
        Gaussian (default): Fits a gaussian pulse to both the time & freq. marginals. Struggles w/ non-bell-shaped spectra.
        Spectrum: Takes the independantly measured spectrum as initial guess with flat phase.
        RANA: Uses "RANA" algorithm to deduce the spectrum from the trace. Uses flat phase. Struggles w/ noise."""
        
    if initialGuess.lower() == 'gaussian':
        if tau is None:
            autocorr =  simpson(trace,wavelengths,axis = 1)
            tau = library_frog.get_FWHM(delays,autocorr)/np.sqrt(2) /np.sqrt(2*np.log(2))
            
            dw = library_frog.get_FWHM(2*np.pi*C/wavelengths[-1::-1],marginal_w[-1::-1])/np.sqrt(2) /np.sqrt(2*np.log(2))
            tau_0 = 2 / (dw)
            
            if tau > tau_0:
                GDD = (tau**2*tau_0**2 - tau_0**4)**0.5/2
            else:
                GDD = 0
        else:
            dw = 2  / (tau/np.sqrt(2*np.log(2))) 
            
        
        w_0 = 2*np.pi*C/lambda_0
        initialGuess = np.complex128(np.exp(- (w_fund-w_0)**2 / dw**2)) * np.exp(1j*GDD*(w_fund-w_0)**2)
    
    elif (initialGuess.lower()=='spectrum') & (marginalCorrection is not None):
       initialGuess = interp( 2*np.pi*C/corrWavelengths[-1::-1], corrSpectrumRaw[-1::-1] ,bounds_error=False, fill_value=0)(w_fund)
       initialGuess[initialGuess<0]=0
       initialGuess = np.complex128(initialGuess**0.5)
       initialGuess /= initialGuess.max()
    else:
        initialGuess = library_frog.RANA(delays,w_shg,trace_w,w_fund)
    
    
    """ Instantiate retriever """
    ret = pypret.Retriever(pnps,method =  method, verbose=True, maxiter=maxIter)


    """ Apply retrieval algorithm and print results """
    ret.retrieve(traceInput, initialGuess)
    results = ret.result()
    
    """ Export retrieved pulse & trace """
    pulseRetrieved = results.pulse_retrieved
    traceRetrieved = results.trace_retrieved
    pulseFrequencies = w_fund/(2*np.pi)
    traceFrequencies = w_shg/(2*np.pi)
    
    """ Make plots """
    axSpectrum = library_frog.plot_output(pulseRetrieved, initialGuess, pulseFrequencies, traceRetrieved, traceFrequencies,delays, wavelengths)
    
    if marginalCorrection is not None:
        axSpectrum.plot(corrWavelengths*1e9,corrSpectrumRaw/corrSpectrumRaw.max(),'g--',linewidth = 3,label = 'Measured')
        axSpectrum.set_ylim([0,1.05])
    axSpectrum.legend()
    
    return pulseRetrieved, initialGuess, pulseFrequencies, traceRetrieved, traceFrequencies,delays, wavelengths



""" Sets the default options for HDF5 data.

Author: Nils Geib, nils.geib@uni-jena.de
"""
import numpy as np
from types import SimpleNamespace as SN


class HDF5Options:
    """ A class that handles the correct HDF5 options for different data sets.

    The reason is simply that native HDF5 compression will actually increase
    the file size for small arrays (< 300 bytes). This class selects different
    HDF5 options based on the dataset over the method ``__call__``. It can
    be subclassed to support more sophisticated selection strategies.
    """
    def __init__(self):
        self.compression_threshold = 300  # bytes
        self.libver = 'latest'
        self.driver = None
        self.kwds = dict()
        self.encoding = 'utf-8'
        self.compressed_dataset = SN(
                compression='gzip', chunks=True, fletcher32=True, shuffle=True,
                compression_opts=9)
        self.dataset = SN(
                compression=None, chunks=None, fletcher32=False, shuffle=False)

    def copy(self):
        ret = HDF5Options()
        ret.compression_threshold = self.compression_threshold
        ret.libver = self.libver
        ret.driver = self.driver
        ret.encoding = self.encoding
        ret.kwds = dict(**self.kwds)
        ret.compressed_dataset = SN(**self.compressed_dataset.__dict__)
        ret.dataset = SN(**self.dataset.__dict__)
        return ret

    def __call__(self, arr):
        ''' Returns the correct dataset creation options for an array.
        '''
        arr = np.asanyarray(arr)
        if arr.nbytes > self.compression_threshold:
            kwargs = self.compressed_dataset.__dict__
        else:
            kwargs = self.dataset.__dict__
        return kwargs


DEFAULT_OPTIONS = HDF5Options()



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 10 13:48:34 2018

@author: Jan
"""

import sys
import numpy as np
import numpy.matlib as matlib
sys.path.append('../')
import femtoQ.plotting as fqp

figure_number = 1
figure_size = (18, 8) # in cm, for presentation 16:9 the size is 33.8:19
axis_pos = fqp.generate_rectangular_axes_pos((.11, .6), (.21))
axis_size = matlib.repmat((.35, .6), 2, 1)
axis_xlim = ([-10, 700], [1,2])
axis_ylim = ([-1.2, 1.5], [0,3])
axis_xlabel = ('time delay $t_d$ (fs)', 'energy (eV)')
axis_ylabel = (r'$\Delta E(h\nu_{max})$ (meV)', 'test')

# %%
fig, ax = fqp.create_axes(figure_number, figure_size, axis_pos, axis_size, axis_xlabel, axis_ylabel, axis_xlim, axis_ylim)

x = np.linspace(-100, 1000, 1000);
y = np.sin(x/200*2*np.pi)

ax[0].plot(x,y)

x2 = np.linspace(1, 2, 100);
y2 = 2*(x2-1)
ax[1].plot(x2,y2)

tax = fqp.create_top_axis(ax[1], 'ev2nm', np.arange(400, 1300, 250), np.arange(400, 1300, 50))

tax[0].set_xlabel('wavelength (nm)')
tax[0].set_xlim(ax[1].get_xlim())
# %%
fig.savefig('test_py.png', dpi=600)
fig.savefig('test_py.pdf', dpi=600)
fig.savefig('test_py.eps', dpi=600)



""" Provides a function to generate random pulses with specified TBP.
"""
import numpy as np
import scipy.optimize as opt
from . import lib


def random_pulse(pulse, tbp, edge_value=None, check=True):
    """ Creates a random pulse with a specified time-bandwidth product.

    Parameters
    ----------
    pulse : Pulse instance
    tbp : float
        The specified time-bandwidth product.
    edge_value : float, optional
        The maximal value for the pulse amplitude at the edges of the grid.
        It defaults to the double value epsilon ~2e-16.

    Returns
    -------
    bool : True on success, False if an error occured. The resulting pulse
        is stored in the Pulse instance passed to the function.

    Notes
    -----
    The function creates random pulses by iteratively restricting the bandwidth
    in time and frequency domain. It starts from random complex values in
    frequency domain, multiplies a Gaussian function, transforms in the
    time domain and multiplies a Gaussian function again. The filter functions
    are Gaussians with the specified time-bandwidth product.
    The TBP of the Gaussian filters, however, does not directly correspond
    to the TBP of the resulting pulse. To use this algorithm to generate a
    pulse with exactly the specified TBP, it is run in the range 0.5 * TBP to
    1.5 * TBP using a scalar root search (brentq). Usually this guarantees
    convergence within a few tries.
    The larger the TBP the larger the number of points has to be. So the
    algorithm may fail to find a solution if pulse.N is too small.
    """
    if edge_value is None:
        # this is roughly the roundoff error induced by an FFT
        edge_value = pulse.N * np.finfo(np.double).eps
    # access/calculate some fundamental grid parameters
    t, w = pulse.t, pulse.w
    t1, t2 = t[0], t[-1]
    w1, w2 = w[0], w[-1]
    t0, w0 = 0.5 * (t1 + t2), 0.5 * (w1 + w2)
    log_edge = np.log(edge_value)

    """ Calculate the width of a Gaussian function that drops exactly to
        edge_value at the edges of the grid.
    """
    spectral_width = np.sqrt(-0.125 * (w1 - w2)**2 / log_edge)
    # Now the same in the temporal domain
    max_temporal_width = np.sqrt(-0.125 * (t1 - t2)**2 / log_edge)
    # The actual temporal width is obtained by the uncertainty relation
    # from the specified TBP
    temporal_width = 2.0 * tbp / spectral_width

    if temporal_width > max_temporal_width:
        print("The required time-bandwidth product cannot be reached! "
              "Decrease edge_value or increase pulse.N!")
        return False

    # special case for TBP = 0.5 (transform-limited case)
    if tbp == 0.5:
        phase = np.exp(1.0j * lib.twopi * np.random.rand(pulse.N))
        pulse.spectrum = lib.gaussian(w, w0, spectral_width) * phase
        return True

    # create the filter functions, the scaling by the number of rounds is
    # purely a heuristic
    spectral_filter = lib.gaussian(w, w0, spectral_width)

    """ The algorithm works by iteratively filtering in the frequency and time
        domain. However, the chosen filter functions only roughly give
        the correct TBP. To obtain the exact result we scale the temporal
        filter bandwidth by a factor and perform a scalar minimization on
        that value.
    """
    spectrum = (np.random.rand(pulse.N) *
                np.exp(1j * lib.twopi * np.random.rand(pulse.N)))
    # rough guess for the relative range in which our optimal value lies
    factor_min, factor_max = 0.5, 1.5

    def create_pulse(factor):
        """ This performs the filtering. """
        temporal_filter = lib.gaussian(t, t0, temporal_width * factor)

        pulse.spectrum = spectrum * spectral_filter
        pulse.field = pulse.field * temporal_filter

    def objective(factor):
        """ This function should be zero """
        create_pulse(factor)
        return tbp - pulse.time_bandwidth_product

    # The objective function has to change sign in the bounds we chose
    i = 0
    while np.sign(objective(factor_min)) == np.sign(objective(factor_max)):
        # for some random arrays this condition is not always fulfilled.
        # just try again
        spectrum = (np.random.rand(pulse.N) *
                    np.exp(1j * lib.twopi * np.random.rand(pulse.N)))
        if i == 10:
            # I have never observed this case.
            raise ValueError('Could not create a pulse for these parameters!')

    # actually perform the optimization
    factor = opt.brentq(objective, factor_min, factor_max)
    # and finally create the pulse
    create_pulse(factor)

    # The random pulse is stored in pulse
    return True


def random_gaussian(pulse, fwhm, phase_max=0.1 * np.pi):
    """ Generates a Gaussian pulse with random phase.

    Its pulse of duration is given by ``fwhm``.
    """
    # convert intensity fwhm to field std-dev.
    sigma = 0.5 * fwhm / np.sqrt(np.log(2.0))
    phase = np.exp(1.0j * np.random.uniform(-phase_max, phase_max, pulse.N))
    pulse.field = lib.gaussian(pulse.t, sigma=sigma) * phase
    return True



""" This module handles conversion between frequency units.

The supported units and their shorthands are:

- wl : wavelength in meter
- om: angular frequency in rad/s
- f: frequency in 1/s
- k: angular wavenumber in rad/m

The conversion functions have the form `shorthand2shorthand` which is not
pythonic but very short. A more pythonic conversion can be achieved by using
the `convert` function

>>> convert(x, 'wl', 'om')

The shorthands will be used throughout the package to identify frequency units.

The functions in this module should be used wherever a frequency convention
is necessary to avoid mistakes and make the code more expressive.
"""
from copy import copy
from .lib import sol, twopi


frequency_labels = {
    'wl': 'wavelength',
    'om': 'angular frequency',
    'f': 'frequency',
    'k': 'angular wavenumber'
}

frequency_units = {
    'wl': 'm',
    'om': 'Hz rad',
    'f': 'Hz',
    'k': 'rad/m'
}


def om2wl(om):
    return twopi/om*sol


def k2wl(k):
    return twopi/k


def f2wl(f):
    return sol/f


def wl2f(wl):
    return sol/wl


def om2f(om):
    return om/twopi


def k2f(k):
    return k*sol/twopi


def wl2om(wl):
    return twopi*sol/wl


def f2om(f):
    return twopi*f


def k2om(k):
    return k*sol


def wl2k(wl):
    return twopi/wl


def om2k(om):
    return om/sol


def f2k(f):
    return twopi*f/sol


# this dictionary can be used for programmatic conversions
conversions = {
    'wl': {
        'wl': lambda x: copy(x),
        'om': wl2om,
        'f': wl2f,
        'k': wl2k
    },
    'om': {
        'wl': om2wl,
        'om': lambda x: copy(x),
        'f': om2f,
        'k': om2k
    },
    'f': {
        'wl': f2wl,
        'om': f2om,
        'f': lambda x: copy(x),
        'k': f2k
    },
    'k': {
        'wl': k2wl,
        'om': k2om,
        'f': k2f,
        'k': lambda x: copy(x)
    }
}


def convert(x, unit1, unit2):
    """ Convert between two frequency units.

    Parameters
    ----------
    x : float or array_like
        Numerical value or array that should be converted.
    unit1, unit2 : str
        Shorthands for the original unit (`unit1`) and the destination unit
        (`unit2`).

    Returns
    -------
    float or array_like
        The converted numerical value or array. It will always be a copy, even
        if `unit1 == unit2`.

    Notes
    -----
    Unit shorthands can be any of
    `wl` : wavelength in meter
    `om` : angular frequency in rad/s
    `f` : frequency in 1/s
    `k` : angular wavenumber in rad/m
    """
    return conversions[unit1][unit2](x)



""" This module implements several helper routines for plotting.
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import EngFormatter
from . import lib
from .frequencies import convert


def plot_meshdata(ax, md, cmap="nipy_spectral"):
    x, y = lib.edges(md.axes[1]), lib.edges(md.axes[0])
    im = ax.pcolormesh(x, y, md.data, cmap=cmap)
    ax.set_xlabel(md.labels[1])
    ax.set_ylabel(md.labels[0])

    fx = EngFormatter(unit=md.units[1])
    ax.xaxis.set_major_formatter(fx)
    fy = EngFormatter(unit=md.units[0])
    ax.yaxis.set_major_formatter(fy)
    return im


class MeshDataPlot:

    def __init__(self, mesh_data, plot=True, **kwargs):
        self.md = mesh_data
        if plot:
            self.plot(**kwargs)

    def plot(self, show=True):
        md = self.md

        fig, ax = plt.subplots()
        im = plot_meshdata(ax, md, "nipy_spectral")
        fig.colorbar(im, ax=ax)

        self.fig, self.ax = fig, ax
        self.im = im
        if show:
            fig.tight_layout()
            plt.show()

    def show(self):
        plt.show()


def plot_complex(x, y, ax, ax2, yaxis='intensity', limit=False,
                 phase_blanking=False, phase_blanking_threshold=1e-3,
                 amplitude_line="r-", phase_line="b-"):
    if yaxis == "intensity":
        amp = lib.abs2(y)
    elif yaxis == "amplitude":
        amp = np.abs(y)
    else:
        raise ValueError("yaxis mode '%s' is unknown!" % yaxis)
    phase = lib.phase(y)
    # center phase by weighted mean
    phase -= lib.mean(phase, amp * amp)
    if phase_blanking:
        x2, phase2 = lib.mask_phase(x, amp, phase, phase_blanking_threshold)
    else:
        x2, phase2 = x, phase
    if limit:
        xlim = lib.limit(x, amp)
        ax.set_xlim(xlim)
        f = (x2 >= xlim[0]) & (x2 <= xlim[1])
        ax2.set_ylim(lib.limit(phase2[f], padding=0.05))

    li1, = ax.plot(x, amp, amplitude_line)
    li2, = ax2.plot(x2, phase2, phase_line)

    return li1, li2, amp, phase


class PulsePlot:

    def __init__(self, pulse, plot=True, **kwargs):
        self.pulse = pulse
        if plot:
            self.plot(**kwargs)

    def plot(self, xaxis='wavelength', yaxis='intensity', limit=True,
             oversampling=False, phase_blanking=False,
             phase_blanking_threshold=1e-3, show=True):
        pulse = self.pulse

        fig, axs = plt.subplots(1, 2)
        ax1, ax2 = axs.flat
        ax12 = ax1.twinx()
        ax22 = ax2.twinx()

        if oversampling:
            t = np.linspace(pulse.t[0], pulse.t[-1], pulse.N * oversampling)
            field = pulse.field_at(t)
        else:
            t = pulse.t
            field = pulse.field

        # time domain
        li11, li12, tamp, tpha = plot_complex(t, field, ax1, ax12, yaxis=yaxis,
                          phase_blanking=phase_blanking, limit=limit,
                          phase_blanking_threshold=phase_blanking_threshold)
        fx = EngFormatter(unit="s")
        ax1.xaxis.set_major_formatter(fx)
        ax1.set_title("time domain")
        ax1.set_xlabel("time")
        ax1.set_ylabel(yaxis)
        ax12.set_ylabel("phase (rad)")

        # frequency domain
        if oversampling:
            w = np.linspace(pulse.w[0], pulse.w[-1], pulse.N * oversampling)
            spectrum = pulse.spectrum_at(w)
        else:
            w = pulse.w
            spectrum = pulse.spectrum

        if xaxis == "wavelength":
            w = convert(w + pulse.w0, "om", "wl")
            unit = "m"
            label = "wavelength"
        elif xaxis == "frequency":
            w = w
            unit = " rad Hz"
            label = "frequency"

        li21, li22, samp, spha = plot_complex(w, spectrum, ax2, ax22, yaxis=yaxis,
                          phase_blanking=phase_blanking, limit=limit,
                          phase_blanking_threshold=phase_blanking_threshold)
        fx = EngFormatter(unit=unit)
        ax2.xaxis.set_major_formatter(fx)
        ax2.set_title("frequency domain")
        ax2.set_xlabel(label)
        ax2.set_ylabel(yaxis)
        ax22.set_ylabel("phase (rad)")

        self.fig = fig
        self.ax1, self.ax2 = ax1, ax2
        self.ax12, self.ax22 = ax12, ax22
        self.li11, self.li12, self.li21, self.li22 = li11, li12, li21, li22
        self.tamp, self.tpha = tamp, tpha
        self.samp, self.spha = samp, spha

        if show:
            fig.tight_layout()
            plt.show()



# -*- coding: utf-8 -*-
"""
Created on Thu Nov 22 16:52:35 2018

@author: Jan
"""
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import rcParams
import numpy as np

def isiterable(p_object):
    try:
        it = iter(p_object)
    except TypeError:
        return False
    return True

def cm2inch(*tupl):
    inch = 2.54
    if isinstance(tupl[0], tuple):
        return tuple(i/inch for i in tupl[0])
    else:
        return tuple(i/inch for i in tupl)


def transform_to_axes_quadruple(axis_pos, axis_size):
    if len(axis_size) < len(axis_pos):
        raise ValueError('The size of not all the axis positions is specified.')

    axis_quad = []
    for i, pos in enumerate(axis_pos):
        axis_quad.append([pos[0], pos[1], axis_size[i][0], axis_size[i][1]])
    return axis_quad


def generate_rectangular_axes_pos(pos_x, pos_y):
    if type(pos_x) is float:
        pos_x = [pos_x]
    else:
        pos_x = list(pos_x)
        pos_x.sort()

    if type(pos_y) is float:
        pos_y = [pos_y]
    else:
        pos_y = list(pos_y)
        pos_y.sort()

    axis_pos = []
    for x in pos_x:
        for y in pos_y:
            axis_pos.append([x, y])
    return axis_pos


def set_default_values_rpl():
    print('Setting default plotting values for RPL')
    plt.rcdefaults()
#   Reopen all you figures once you changed something here!
    rcParams['font.family'] = 'serif'
    rcParams['font.sans-serif'] = ['Tahoma']
    rcParams['font.serif'] = ['Times New Roman']
    rcParams['font.size'] = 11
    rcParams['axes.linewidth'] = 1
    rcParams['figure.dpi'] = 200
    rcParams['lines.linewidth'] = 1
    rcParams['xtick.top'] = True
    rcParams['ytick.right'] = True
    rcParams['xtick.direction'] = 'in'
    rcParams['ytick.direction'] = 'in'
    rcParams['xtick.minor.visible'] = True
    rcParams['ytick.minor.visible'] = True
    rcParams['xtick.major.size'] = 5
    rcParams['xtick.minor.size'] = 3
    rcParams['xtick.major.width'] = 1
    rcParams['xtick.minor.width'] = 1
    rcParams['ytick.major.size'] = rcParams['xtick.major.size']
    rcParams['ytick.minor.size'] = rcParams['xtick.minor.size']
    rcParams['ytick.major.width'] = rcParams['xtick.major.width']
    rcParams['ytick.minor.width'] = rcParams['xtick.minor.width']
    rcParams['mathtext.default'] = 'regular'
    rcParams["legend.frameon"] = False
    rcParams["legend.handlelength"] = 1
    rcParams["legend.handletextpad"] = .5
    rcParams["legend.columnspacing"] = .7

def set_default_values_cleo():
    print('Setting default plotting values for CLEO')
    plt.rcdefaults()
#   Reopen all you figures once you changed something here!
    rcParams['font.family'] = 'serif'
    rcParams['font.sans-serif'] = ['Tahoma']
    rcParams['font.serif'] = ['Times New Roman']
    rcParams['font.size'] = 9
    rcParams['axes.linewidth'] = 1
    rcParams['figure.dpi'] = 200
    rcParams['lines.linewidth'] = 1
    rcParams['xtick.top'] = True
    rcParams['ytick.right'] = True
    rcParams['xtick.direction'] = 'in'
    rcParams['ytick.direction'] = 'in'
    rcParams['xtick.minor.visible'] = True
    rcParams['ytick.minor.visible'] = True
    rcParams['xtick.major.size'] = 5
    rcParams['xtick.minor.size'] = 3
    rcParams['xtick.major.width'] = 1
    rcParams['xtick.minor.width'] = 1
    rcParams['ytick.major.size'] = rcParams['xtick.major.size']
    rcParams['ytick.minor.size'] = rcParams['xtick.minor.size']
    rcParams['ytick.major.width'] = rcParams['xtick.major.width']
    rcParams['ytick.minor.width'] = rcParams['xtick.minor.width']
    rcParams['mathtext.rm'] = 'serif'
    rcParams['mathtext.default'] = 'regular'
    rcParams['mathtext.fontset'] = 'dejavuserif'
    rcParams["legend.frameon"] = False
    rcParams["legend.handlelength"] = 1
    rcParams["legend.handletextpad"] = .5
    rcParams["legend.columnspacing"] = .7

def set_default_values_presentation():
    print('Setting default plotting values for presentations')
    plt.rcdefaults()
#   Reopen all you figures once you changed something here!
    rcParams['font.family'] = 'sans-serif'
    rcParams['font.sans-serif'] = ['Tahoma']
    rcParams['font.serif'] = ['Times New Roman']
    rcParams['font.size'] = 18
    rcParams['axes.linewidth'] = 2
    #rcParams['figure.dpi'] = 100
    rcParams['lines.linewidth'] = 2
    rcParams['xtick.top'] = True
    rcParams['ytick.right'] = True
    rcParams['xtick.direction'] = 'in'
    rcParams['ytick.direction'] = 'in'
    rcParams['xtick.minor.visible'] = True
    rcParams['ytick.minor.visible'] = True
    rcParams['xtick.major.size'] = 7
    rcParams['xtick.minor.size'] = 5
    rcParams['xtick.major.width'] = 2
    rcParams['xtick.minor.width'] = rcParams['xtick.major.width']
    rcParams['ytick.major.size'] = rcParams['xtick.major.size']
    rcParams['ytick.minor.size'] = rcParams['xtick.minor.size']
    rcParams['ytick.major.width'] = rcParams['xtick.major.width']
    rcParams['ytick.minor.width'] = rcParams['xtick.minor.width']
    rcParams['mathtext.default'] = 'regular'
    rcParams['mathtext.fontset'] = 'dejavuserif'
    rcParams["legend.frameon"] = False
    rcParams["legend.handlelength"] = 1
    rcParams["legend.handletextpad"] = .5
    rcParams["legend.columnspacing"] = .7
    rcParams['figure.figsize'] = 6*1.4, 6

def create_axes(figure_number, figure_size, axis_pos, axis_size, axis_xlabel, axis_ylabel, axis_xlim, axis_ylim, label_offset = 0):
    set_default_values_presentation()
    fig = plt.figure(figure_number)
    fig.clear()
    fig.set_size_inches(cm2inch(figure_size), forward=True)

    axis_quad = transform_to_axes_quadruple(axis_pos, axis_size)

    ax = []
    for i, item in enumerate(axis_quad):
        ax.append(fig.add_axes(item))

        if(type(axis_xlabel) is str and i == 0):
            ax[-1].set_xlabel(axis_xlabel)
        elif(type(axis_xlabel) is not str and len(axis_xlabel) > i):
            ax[-1].set_xlabel(axis_xlabel[i])

        if(type(axis_ylabel) is str and i == 0):
            ax[-1].set_ylabel(axis_ylabel)
        elif(type(axis_ylabel) is not str and len(axis_xlabel) > i):
            ax[-1].set_ylabel(axis_ylabel[i])

#       Empty label removes also ticklabels
        if(ax[-1].get_xlabel() == ''):
            ax[-1].tick_params(labelbottom=False)
        if(ax[-1].get_ylabel() == ''):
            ax[-1].tick_params(labelleft=False)

        if(len(axis_xlim) > i):
            if(isiterable(axis_xlim[i]) and len(axis_xlim[i]) == 2):
                ax[-1].set_xlim(axis_xlim[i], auto=False)
            elif i == 0:
                ax[-1].set_xlim(axis_xlim, auto=False)

        if(len(axis_ylim) > i):
            if(isiterable(axis_ylim[i]) and len(axis_ylim[i]) == 2):
                ax[-1].set_ylim(axis_ylim[i], auto=False)
            elif i == 0:
                ax[-1].set_ylim(axis_ylim, auto=False)
            ax[-1].autoscale(False)

        if(len(axis_quad) > 1):
            ax[-1].annotate("("+chr(97+i+label_offset)+")",(.05, .85), xycoords='axes fraction')


    ax = tuple(ax)
    return(fig, ax)


def create_top_axis(axes, conversion, major, minor):
    if type(axes) is mpl.axes._axes.Axes:
        axes = (axes, ) #make it a tuple so indexing works

    tax = [None] * len(axes)
    for i, ax in enumerate(axes):
        tax[i] = ax.twiny()
        for axis in ['top', 'bottom', 'left', 'right']:
            tax[i].spines[axis].set_linewidth(ax.spines[axis].get_linewidth())
        if conversion == 'ev2nm':
            new_major = 1240./major
            new_minor = 1240./minor

        ind = np.argsort(new_major)
        new_major = new_major[ind]
        major = major[ind]
        new_minor.sort()
        tax[i].set_xticks(new_minor, minor=True)
        tax[i].set_xticks(new_major)
        tax[i].set_xticklabels(list(map(str, major)))
#        This needs to be at the end, at least after setting the ticks!
        tax[i].autoscale(False)
        tax[i].set_xlim(ax.get_xlim())
    return tax

def create_side_axis(axes, conversion, major, minor):
    if type(axes) is mpl.axes._axes.Axes:
        axes = (axes, ) #make it a tuple so indexing works

    tax = [None] * len(axes)
    for i, ax in enumerate(axes):
        tax[i] = ax.twinx()
        for axis in ['top', 'bottom', 'left', 'right']:
            tax[i].spines[axis].set_linewidth(ax.spines[axis].get_linewidth())
        if conversion == 'ev2nm':
            new_major = 1240./major
            new_minor = 1240./minor

        ind = np.argsort(new_major)
        new_major = new_major[ind]
        major = major[ind]
        new_minor.sort()
        tax[i].set_yticks(new_minor, minor=True)
        tax[i].set_yticks(new_major)
        tax[i].set_yticklabels(list(map(str, major)))
#        This needs to be at the end, at least after setting the ticks!
        tax[i].autoscale(False)
        tax[i].set_xlim(ax.get_xlim())
    return tax



""" This module implements testing procedures for retrieval algorithms.
"""
import numpy as np
from scipy import optimize
from . import lib


def pulse_error(E, E0, ft, dot_ambiguity=False,
                spectral_shift_ambiguity=False):
    ''' Calculates the normalized rms error between two pulse spectra while
    taking into account the retrieval ambiguities.

    One step in `optimal_rms_error` (the determination of the initial bracket)
    could probably be more efficient, see [Dorrer2002]_). We use the less
    elegant but maybe more straightforward way of simply sampling the
    range for a bracket that encloses a minimum.

    Parameters
    ----------
    E, E0: 1d-array
        Complex-valued arrays that contain the spectra of the pulses.
        ``E`` will be matched against ``E0``.
    ft : FourierTransform instance
        Performs Fourier transforms on the pulse grid.
    dot_ambiguity : bool, optional
        Takes the direction of time ambiguity into account. Default is
        ``False``.
    spectral_shift_ambiguity : bool, optional
        Takes the spectral shift ambiguity into account. Default is ``False``.
    '''
    test_fields = [[ft.w, E, E0]]
    if spectral_shift_ambiguity:
        # spectrally shift by exactly half the grid size
        Et = ft.backward(E)
        Et *= np.exp(0.5j * ft.N * ft.dw * ft.t)
        test_fields.append([ft.w, ft.forward(Et), E0])
    if dot_ambiguity:
        max_iter = len(test_fields)
        for i in range(max_iter):
            tf = test_fields[i]
            test_fields.append([tf[0], tf[1].conj(), tf[2]])

    best_error = np.inf
    for w, spec1, spec2 in test_fields:
        error, matched = optimal_rms_error(w, spec1, spec2)
        if error < best_error:
            best_error = error
            best_match = matched
    return best_error, best_match


def best_constant_phase(E, E0):
    """ Finds ``c`` with ``|c| = 1`` so that ``sum(abs2(c * y1 - y2))`` is
    minimal.

    Uses an analytic solution.
    """
    A = np.sum(E.conj() * E0)
    c = A / np.abs(A)
    err1 = np.sum(lib.abs2(c * E - E0))
    err2 = np.sum(lib.abs2(-c * E - E0))
    if err2 < err1:
        c = -c
    return c


def optimal_rms_error(w, E, E0):
    """ Calculates the RMS error of two arrays, ignoring scaling, constant
    and linear phase of one of them.

    Formally it calculates the minimal error::

        R = sqrt(|rho * exp(i*(x*a + b)) * y1 - y2|^2 / |y2|^2)

    with respect to rho, a and b. If additionally ``conjugation = True`` then
    the error for conjugate(y1) is calculated and the best transformation of y1
    is also returned.
    """
    # E is rescaled so that the amplitudes match in the least-squares sense
    E = E * lib.best_scale(E, E0)

    # find optimal linear and constant phase
    # determine the frequency spacing
    dw = np.max(np.abs(np.diff(w)))
    # rescale the objective function to make it easier for the optimizer
    scale = 1.0 / np.sqrt(np.sum(lib.abs2(E0)) * E.shape[0])

    def objective(alpha):
        linear = np.exp(1.0j * alpha / dw * w)
        phase0 = best_constant_phase(linear * E, E0)
        cresiduals = (phase0 * linear * E - E0) * scale
        return lib.norm2(cresiduals)

    # find an initial bracket
    alphas = np.linspace(-np.pi, np.pi, 2 * E.shape[0])
    err = np.array([objective(a) for a in alphas])
    idx = np.argmin(err)
    bracket = [
        alphas[max(0, idx - 1)],
        alphas[min(alphas.shape[0] - 1, idx + 1)]
    ]
    # run a bounded optimization on that bracket to obtain high precision
    res = optimize.minimize_scalar(
        objective,
        bounds=bracket,
        method='bounded',
        options=dict(maxiter=100, xatol=1e-10)
    )
    linear = np.exp(1.0j * res.x / dw * w)
    phase0 = best_constant_phase(linear * E, E0)
    E = phase0 * linear * E

    return lib.nrms(E, E0), E



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="femtoMod",
    version="2.1.9",
    author="Patrick, Benjamin, Etienne",
    description="femtoQ Modules",
    long_description=long_description,
    url="https://github.com/UltraFastQ/Modules",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
from scipy.constants import c as C
from scipy.interpolate import interp1d as interp
from scipy.integrate import simps as simpson
import matplotlib.pyplot as plt
import femtoQ.tools as fq



def get_FWHM(t,I_t):
    
    tmp = np.linspace(t[0],t[-1],1000000)
    
    I_tmp = interp(t,I_t,'quadratic')
    I_tmp = I_tmp(tmp)
    
    I_tmp = I_tmp - np.min(I_tmp)
    I_tmp /= np.max(I_tmp)

    t_1 = tmp[I_tmp>=0.5][0]
    t_2 = tmp[I_tmp>=0.5][-1]

    return t_2-t_1

def unpack_data(filename,wavelengthLimits):
    
    data = np.load(filename)
    
    delays = data['time']*1e-15
    wavelengthsSpectro = data['wavelengths']*1e-9
    trace = data['trace']
    trace/= np.max(trace)
    
    trace = trace[:,((wavelengthsSpectro>wavelengthLimits[0])&(wavelengthsSpectro<wavelengthLimits[1]))]
    wavelengthsSpectro = wavelengthsSpectro[((wavelengthsSpectro>wavelengthLimits[0])&(wavelengthsSpectro<wavelengthLimits[1]))]
    
    data.close()
    
    

    plt.figure()
    plt.pcolormesh(delays*1e15,wavelengthsSpectro*1e9,trace.transpose())
    plt.title('Measured trace')
    plt.ylabel('Wavelengths [nm]')
    plt.xlabel('Delay [fs]')
    plt.colorbar()


    
    return delays, wavelengthsSpectro, trace


def plot_output(pulseRetrieved, initialGuess, pulseFrequencies, traceRetrieved, traceFrequencies,delays, wavelengthsSpectro):

    t, E = freq2time(pulseFrequencies, pulseRetrieved)
    
    t, E_cmp = freq2time(pulseFrequencies, np.abs(pulseRetrieved))

    t0 = t[np.argmax(np.abs(E)**2)]
    E = interp(t-t0,E,'quadratic',bounds_error=False,fill_value=0)
    E = E(t)
    
    print('FWHM duration: ' + str( round(get_FWHM(t, np.abs(E)**2)*1e15,1) ) + ' fs')
    print('FWHM duration: ' + str( round(get_FWHM(t, np.abs(E_cmp)**2)*1e15,1) ) + ' fs')
    
    plt.figure()
    plt.plot(t*1e15,np.abs(E)**2 / np.max(np.abs(E)**2),'b',label = 'Retrieved')
    plt.plot(t*1e15,np.abs(E_cmp)**2 / np.max(np.abs(E_cmp)**2),'k',label = 'Transform limit')
    plt.xlabel('Time [fs]')
    plt.ylabel('Power [arb. u.]')
    plt.legend()
    
    
    wavelengths = C/(pulseFrequencies)
    II = np.argsort(wavelengths)
    
    IIplotphase = ( np.abs(pulseRetrieved[II])**2 / np.max(np.abs(pulseRetrieved[II])**2) ) > 0.1
    
    plt.figure()
    axL = plt.gca()
    axR = axL.twinx()
    axL.plot(wavelengths[II]*1e9,np.abs(initialGuess[II])**2,'r',linewidth = 3,label = 'Initial guess')
    axL.plot(wavelengths[II]*1e9,np.abs(pulseRetrieved[II])**2 / np.max(np.abs(pulseRetrieved[II])**2),'k',linewidth = 3,label = 'Retrieved spectrum')
    axL.set_ylabel('Normalized power density')
    axL.set_xlabel('Wavelengths [nm]')
    axR.set_ylabel('Spectral phase (x $\pi$) [rad]')
    
    retrievedPhase = np.unwrap(np.angle(pulseRetrieved))
    retrievedPhase -= np.average(retrievedPhase[II],weights = np.abs(pulseRetrieved[II])**2)
    
    axR.plot(wavelengths[II][IIplotphase]*1e9,retrievedPhase[II][IIplotphase]/np.pi,'--k')
    axR.set_ylim(retrievedPhase[II][IIplotphase].min()-np.abs(retrievedPhase[II][IIplotphase].min()*0.1)/np.pi,retrievedPhase[II][IIplotphase].max()*1.1/np.pi)
    plt.xlim(wavelengthsSpectro[0]*1.8e9,wavelengthsSpectro[-1]*2.2e9)
    
    plt.figure()
    plt.pcolormesh(delays*1e15,(C/traceFrequencies)*1e9,traceRetrieved.transpose())
    plt.title('Retrieved trace')
    plt.ylabel('Wavelengths [nm]')
    plt.xlabel('Delay [fs]')
    plt.ylim(wavelengthsSpectro[0]*1e9,wavelengthsSpectro[-1]*1e9)
    plt.colorbar()

    return axL


def freq2time(frequency, spectrum):
    # Interpolate over new frequency grid, including negative w components
    v_max = frequency[-1]
    v_min = -v_max
    N = len(spectrum)
    dv = v_max/N
    
    new_v = np.hstack((np.array([0]),np.linspace(dv,v_max,N),np.linspace(v_min,-dv,N)))
    
    interp_marginal = interp(frequency,spectrum,'quadratic',bounds_error=False,fill_value=0)
    spectrum = interp_marginal(new_v)

    E = np.fft.fftshift(np.fft.ifft(spectrum))
    t = np.fft.fftshift(np.fft.fftfreq(2*N+1,dv))

    return t, E

def RANA(delays,w_shg,trace_w,w_fund):
    
    a = 0.09
    b = 0.425
    c = 0.1
    
    # Integrate over delay axis
    marginal_w = simpson(trace_w,delays,axis = 0)
    
    marginal_w = fq.ezsmooth(marginal_w,11,'hanning')
    
    marginal_w[marginal_w<0] = 0
    
    # Interpolate over new frequency grid, including negative w components
    w_max = w_shg[-1]*2
    w_min = -w_max
    N = len(marginal_w)*2
    dw = w_max/N
    
    new_w = np.hstack((np.array([0]),np.linspace(dw,w_max,N),np.linspace(w_min,-dw,N)))
    
    interp_marginal = interp(w_shg,marginal_w,'quadratic',bounds_error=False,fill_value=0)
    marginal_w = interp_marginal(new_w)

    S = np.fft.fftshift(np.fft.ifft(marginal_w))
    t = np.fft.fftshift(np.fft.fftfreq(2*N+1,dw/2/np.pi))

    s_p = np.sqrt(S)
    s_m = -np.sqrt(S)
    
    s = np.zeros_like(s_p)
    for ii,tau in enumerate(t):
        
        if ii==0:
            s[ii] = s_m[ii]
            
        if ii==1:
            ds0_p = abs(s_p[ii] - s[ii-1])
            ds0_m = abs(s_m[ii] - s[ii-1])
            
            ds1_p = 0
            ds1_m = 0
            
            ds2_p = 0
            ds2_m = 0
            
        
        if ii==2:
            ds0_p = abs(s_p[ii] - s[ii-1])
            ds0_m = abs(s_m[ii] - s[ii-1])
            
            ds1_p = abs( ds0_p - (s[ii-1] - s[ii-2]) )
            ds1_m = abs( ds0_m - (s[ii-1] - s[ii-2]) )
        
            ds2_p = 0
            ds2_m = 0
        
        else:
            ds0_p = abs(s_p[ii] - s[ii-1])
            ds0_m = abs(s_m[ii] - s[ii-1])
            
            ds1_p = abs( ds0_p - (s[ii-1] - s[ii-2]) )
            ds1_m = abs( ds0_m - (s[ii-1] - s[ii-2]) )
            
            ds2_p = abs( ds1_p - ( (s[ii-1] - s[ii-2]) - (s[ii-2] - s[ii-3]) ) )
            ds2_m = abs( ds1_m - ( (s[ii-1] - s[ii-2]) - (s[ii-2] - s[ii-3]) ) )
            
        e_p = a*ds0_p**2 + b*ds1_p**2 + c*ds2_p**2
        e_m = a*ds0_m**2 + b*ds1_m**2 + c*ds2_m**2
        
        
        if e_p < e_m:
            s[ii] = s_p[ii]
        else:
            s[ii] = s_m[ii]
    
    spectrum = np.fft.fftshift(np.fft.fft(s*np.hanning(len(s))))
    v = np.fft.fftshift((np.fft.fftfreq(2*N+1,t[1]-t[0])))
    spectrum = np.abs(spectrum)**0.5
    spectrum_interp = interp(v*2*np.pi,spectrum,'quadratic',bounds_error=False,fill_value=0)
    spectrum = spectrum_interp(w_fund)
    spectrum -= np.min(spectrum)
    spectrum /= np.max(spectrum)
    
    
    spectrum = np.complex128(spectrum) #* np.exp(1j*np.random.normal(0,np.pi/4,len(spectrum)))
    return spectrum


""" Implements functions that handle the serialization of types and classes.

Type handlers store and load objects of exactly that type. Instance handlers
work also work for subclasses of that type.

The instance handlers are processed in the order they are stored. This means
that if an object is an instance of several handled classes it will not raise
an error and will be handled by the first matching handler in the OrderedDict.
"""
import numpy as np
import types
import inspect
from collections import OrderedDict

""" The handler dictionaries are automatically filled when Handler class
    definitions are parsed via the metaclass __new__ function.
"""
# Saver dictionaries: classes as keys, handler classes as values
type_saver_handlers = dict()
instance_saver_handlers = OrderedDict()
# Loader dictionaries: class names as keys, handler classes as values
loader_handlers = dict()


def classname(val):
    """ Returns a qualified class name as string.

    The qualified class name consists of the module and the class name,
    separated by a dot. If an instance is passed to this function, the name
    of its class is returned.

    Parameters
    ----------
    val : instance or class
        The instance or a class of which the qualified class name is returned.

    Returns
    -------
    str : The qualified class name.
    """
    if inspect.isclass(val):
        return ".".join([val.__module__,
                         val.__name__])
    return ".".join([val.__class__.__module__,
                     val.__class__.__name__])


def set_attribute(level, key, value):
    level.attrs[key] = np.string_(value)


def get_attribute(level, key):
    return level.attrs[key].decode('ascii')


def set_classname(level, clsname):
    set_attribute(level, '__class__', clsname)


def get_classname(level):
    return get_attribute(level, '__class__')


def save_to_level(val, level, options, name=None):
    """ A generic save function that dispatches the correct handler.
    """
    t = type(val)
    if t in type_saver_handlers:
        return type_saver_handlers[t].save_to_level(val, level, options, name)
    for i in instance_saver_handlers:
        if isinstance(val, i):
            return instance_saver_handlers[i].save_to_level(val, level,
                                                            options, name)
    raise ValueError("%s of type %s is not supported by any handler!" %
                     (str(val), str(t)))


def load_from_level(level, obj=None):
    """ Loads an object from an HDF5 group or dataset.

    Parameters
    ----------
    level : h5py.Dataset or h5py.Group
        An HDF5 node that stores an object in a valid format.
    obj : instance or None
        If provided this instance will be updated from the HDF5 node instead
        of creating a new instance of the stored object.

    Returns
    -------
    instance of the stored object
    """
    clsname = get_classname(level)
    if clsname not in loader_handlers:
        raise ValueError('Class `%s` has no registered handler.' % clsname)
    handler = loader_handlers[clsname]
    return handler.load_from_level(level, obj=obj)


class TypeRegister(type):
    """ Metaclass that registers a type handler in a global dictionary.
    """
    def __new__(cls, clsname, bases, attrs):
        # convert all methods to classmethods
        for attr_name, attr_value in attrs.items():
            if isinstance(attr_value, types.FunctionType):
                attrs[attr_name] = classmethod(attr_value)
        newclass = super().__new__(cls, clsname, bases, attrs)
        # register the class as a handler for all specified types
        for t in newclass.types:
            newclass.register(t)
        return newclass


class InstanceRegister(type):
    """Metaclass that registers an instance handler in global dictionary.
    """
    def __new__(cls, clsname, bases, attrs):
        # convert all methods to classmethods
        for attr_name, attr_value in attrs.items():
            if isinstance(attr_value, types.FunctionType):
                attrs[attr_name] = classmethod(attr_value)
        newclass = super().__new__(cls, clsname, bases, attrs)
        # register the class as a handler for all specified instances
        for t in newclass.instances:
            newclass.register(t)
        return newclass


class Handler:
    # a default for subclasses (only `group` subclasses have to overwrite)
    level_type = 'dataset'

    @classmethod
    def save_to_level(cls, val, level, options, name):
        """ A generic wrapper around the custom save method that each
            handler implements. It creates a dataset or a group depending
            on the `level_type` class attribute and sets the `__class__`
            attribute correctly.
            For more flexibility subclasses can overwrite this method.
        """
        # get the qualified class name of the object to be saved
        clsname = classname(val)
        # create the dataset or the group and call the save method on it
        if cls.is_dataset() and name is None:
            # if we want to save a dataset in the root group (name = None)
            # we have to give it a name
            name = "default"
        if cls.is_group():
            if name is not None:
                level = cls.create_group(level, name, options)
            set_classname(level, clsname)
        ret = cls.save(val, level, options, name)
        if cls.is_dataset():
            set_classname(ret, clsname)
        return ret

    @classmethod
    def load_from_level(cls, level, obj=None):
        """ The loader that has to be implemented by subclasses.
        """
        raise NotImplementedError()

    @classmethod
    def create_group(cls, level, name, options):
        return level.create_group(name)

    @classmethod
    def create_dataset(cls, data, level, name, **kwargs):
        ds = level.create_dataset(name, data=data, **kwargs)
        return ds

    @classmethod
    def is_group(cls):
        return cls.level_type == 'group'

    @classmethod
    def is_dataset(cls):
        return cls.level_type == 'dataset'

    @classmethod
    def get_type(cls, level):
        return cls.casting[get_classname(level)]


class TypeHandler(Handler, metaclass=TypeRegister):
    """ Handles data of a specific type or class.
    """
    types = []
    casting = {}

    @classmethod
    def register(cls, t):
        global type_saver_handlers, loader_handlers
        if t in type_saver_handlers:
            raise ValueError('Type `%s` is already handled by `%s`.' %
                             (str(t), str(type_saver_handlers[t])))
        typename = classname(t)
        type_saver_handlers[t] = cls
        loader_handlers[typename] = cls
        cls.casting[typename] = t


class InstanceHandler(Handler, metaclass=InstanceRegister):
    """ Handles all instances of a specific (parent) class.

    If an instance is subclass to several classes for which a handler exists,
    no error will be raised (in contrast to TypeHandler). Rather, the first
    match in the global instance_saver_handlers OrderedDict will be used.
    """
    instances = []
    casting = {}

    @classmethod
    def register(cls, t):
        global instance_saver_handlers, loader_handlers
        if t in instance_saver_handlers:
            raise ValueError('Instance `%s` is already handled by `%s`.' %
                             (str(t), str(instance_saver_handlers[t])))
        typename = classname(t)
        instance_saver_handlers[t] = cls
        loader_handlers[typename] = cls
        cls.casting[typename] = t


# Specific handlers
class NoneHandler(TypeHandler):
    types = [type(None)]

    def save(cls, val, level, options, name):
        ds = cls.create_dataset(0, level, name, **options(0))
        return ds

    def load_from_level(cls, level, obj=None):
        return None


class ScalarHandler(TypeHandler):
    types = [float, bool, complex, np.int8, np.int16, np.int32,
             np.int64, np.uint8, np.uint16, np.uint32, np.uint64,
             np.float16, np.float32, np.float64, np.bool_, np.complex64,
             np.complex128]

    def save(cls, val, level, options, name):
        ds = cls.create_dataset(val, level, name, **options(val))
        return ds

    def load_from_level(cls, level, obj=None):
        # cast to the correct type
        type_ = cls.get_type(level)
        # retrieve scalar dataset
        return type_(level[()])


class IntHandler(TypeHandler):
    """ Special int handler to deal with Python's variable size ints.

    They are stored as byte arrays. Probably not the most efficient solution...
    """
    types = [int]

    def save(cls, val, level, options, name):
        val = val.to_bytes((val.bit_length() + 7) // 8, byteorder='little')
        data = np.frombuffer(val, dtype=np.uint8)
        ds = cls.create_dataset(data, level, name, **options(data))
        return ds

    def load_from_level(cls, level, obj=None):
        return int.from_bytes(level[:].tobytes(), byteorder='little')


class TimeHandler(TypeHandler):
    types = [np.datetime64, np.timedelta64]

    def save(cls, val, level, options, name):
        val2 = val.view('<i8')
        ds = cls.create_dataset(val2, level, name, **options(val2))
        set_attribute(ds, '__dtype__', val.dtype)
        return ds

    def load_from_level(cls, level, obj=None):
        val = level[()]
        dtype = get_attribute(level, '__dtype__')
        return val.view(dtype)


class StringHandler(TypeHandler):
    types = [str]

    def save(cls, val, level, options, name):
        b = val.encode(encoding=options.encoding)
        ds = BytesHandler.save(b, level, options, name)
        set_attribute(ds, '__encoding__', options.encoding)
        return ds

    def load_from_level(cls, level, obj=None):
        bstring = level[:].tobytes()
        return bstring.decode(get_attribute(level, '__encoding__'))


class BytesHandler(TypeHandler):
    types = [bytes]

    def save(cls, val, level, options, name):
        data = np.frombuffer(val, dtype=np.uint8)
        ds = cls.create_dataset(data, level, name, **options(data))
        return ds

    def load_from_level(cls, level, obj=None):
        return level[:].tobytes()


class DictHandler(TypeHandler):
    level_type = 'group'
    types = [dict]

    def save(cls, val, level, options, name):
        for key, value in val.items():
            save_to_level(value, level, options, key)

    def load_from_level(cls, level, obj=None):
        obj = dict()
        for key, value in level.items():
            obj[key] = load_from_level(value)
        return obj


class SimpleNamespaceHandler(TypeHandler):
    level_type = 'group'
    types = [types.SimpleNamespace]

    def save(cls, val, level, options, name):
        for key, value in val.__dict__.items():
            save_to_level(value, level, options, key)

    def load_from_level(cls, level, obj=None):
        obj = types.SimpleNamespace()
        for key, value in level.items():
            setattr(obj, key, load_from_level(value))
        return obj


class ListHandler(TypeHandler):
    """ Despite its name it also handles tuples.
    """
    level_type = 'group'
    types = [list, tuple]

    def save(cls, val, level, options, name):
        for idx, element in enumerate(val):
            save_to_level(element, level, options, 'idx_%d' % idx)

    def load_from_level(cls, level, obj=None):
        obj = []
        length = len(list(level.keys()))
        for idx in range(length):
            obj.append(load_from_level(level['idx_%d' % idx]))
        # cast to tuple if necessary
        type_ = cls.get_type(level)
        return type_(obj)


class NDArrayHandler(TypeHandler):
    types = [np.ndarray]

    def save(cls, val, level, options, name):
        ds = cls.create_dataset(val, level, name, **options(val))
        return ds

    def load_from_level(cls, level, obj=None):
        return level[:]



""" Miscellaneous helper functions

These functions fulfill small numerical tasks used in several places in the
package.
"""
import numpy as np
# make numba jit an optional dependence
# see https://github.com/numba/numba/issues/3735
try:
    from numba import jit
except ImportError:
    def jit(pyfunc=None, **kwargs):
        def wrap(func):
            return func
        if pyfunc is not None:
            return wrap(pyfunc)
        else:
            return wrap

# Constants for convenience (not more accurate)
# two pi
twopi = 6.2831853071795862
# sqrt(2)
sqrt2 = 1.4142135623730951
# speed of light
sol = 299792458.0

# Constants that give slightly more accuracy (~1 ulp)
# sqrt(2 * pi)
sqrt2pi = 2.5066282746310007


def as_list(x):
    """ Try to convert argument to list and return it.

    Useful to implement function arguments that could be scalar values
    or lists.
    """
    try:
        return list(x)
    except TypeError:
        return list([x])

@jit(nopython=True, cache=True)
def abs2(x):
    """ Calculates the squared magnitude of a complex array.
    """
    return x.real * x.real + x.imag * x.imag


def rms(x, y):
    """ Calculates the root mean square (rms) error between ``x`` and ``y``.
    """
    return np.sqrt(abs2(x - y).mean())


@jit(nopython=True, cache=True)
def norm2(x):
    """ Calculates the squared L2 or Euclidian norm of array ``x``.
    """
    return abs2(x).sum()


@jit(nopython=True, cache=True)
def norm(x):
    """ Calculates the L2 or Euclidian norm of array ``x``.
    """
    return np.sqrt(abs2(x).sum())


def phase(x):
    """ The phase of a complex array."""
    return np.unwrap(np.angle(x))


def nrms(x, y):
    """ Calculates the normalized rms error between ``x`` and ``y``.

    The convention for normalization varies. Here we use::

        max |y|

    as normalization.
    """
    n = np.abs(y).max()
    if n == 0.0:
        raise ValueError("Second array cannot be zero.")
    return rms(x, y) / n


def mean(x, y):
    """ Calculates the mean of the distribution described by (x, y).
    """
    return np.sum(x * y) / np.sum(y)


def variance(x, y):
    """ Calculates the variance of the distribution described by (x, y).
    """
    dx = x - mean(x, y)
    return np.sum(dx * dx * y) / np.sum(y)


def standard_deviation(x, y):
    """ Calculates the standard deviation of the distribution described by
        (x, y).
    """
    return np.sqrt(variance(x, y))


def gaussian(x, x0=0.0, sigma=1.0):
    """ Calculates a Gaussian function with center ``x0`` and standard
        deviation ``sigma``.
    """
    d = (x - x0) / sigma
    return np.exp(-0.5 * d * d)


def rescale(x, window=[0.0, 1.0]):
    """ Rescales a numpy array to the range specified by ``window``.

    Default is [0, 1].
    """
    maxx = np.max(x)
    minx = np.min(x)
    return (x - minx) / (maxx - minx) * (window[1] - window[0]) + window[0]


def marginals(data, normalize=False, axes=None):
    """ Calculates the marginals of the data array.

    axes specifies the axes of the marginals, e.g., the axes on which the
    sum is projected.

    If axis is None a list of all marginals is returned.
    """
    if axes is None:
        axes = range(data.ndim)
    axes = as_list(axes)
    full_axes = list(range(data.ndim))
    m = []
    for i in axes:
        # for the marginal sum over all axes except the specified one
        margin_axes = tuple(j for j in full_axes if j != i)
        m.append(np.sum(data, axis=margin_axes))
    if normalize:
        m = [rescale(mx) for mx in m]
    return tuple(m) if len(m) != 1 else m[0]


def find(x, condition, n=1):
    """ Return the index of the nth element that fulfills the condition.
    """
    search_n = 1
    for i in range(len(x)):
        if condition(x[i]):
            if search_n == n:
                return i
            search_n += 1
    return -1


def best_scale(E, E0):
    """ Scales rho so that::

        sum (rho * |E| - |E0|)^2

    is minimal.
    """
    Eabs, E0abs = np.abs(E), np.abs(E0)
    return np.sum(Eabs * E0abs) / np.sum(Eabs * Eabs)


def arglimit(y, threshold=1e-3, padding=0.0, normalize=True):
    """ Returns the first and last index where `y >= threshold * max(abs(y))`.
    """
    t = np.abs(y)
    if normalize:
        t /= np.max(t)

    idx1 = find(t, lambda x: x >= threshold)
    if idx1 == -1:
        idx1 = 0
    idx2 = find(t[::-1], lambda x: x >= threshold)
    if idx2 == -1:
        idx2 = t.shape[0] - 1
    else:
        idx2 = t.shape[0] - 1 - idx2

    return (idx1, idx2)


def limit(x, y=None, threshold=1e-3, padding=0.25, extend=True):
    """ Returns the maximum x-range where the y-values are sufficiently large.

    Parameters
    ----------
    x : array_like
        The x values of the graph.
    y : array_like, optional
        The y values of the graph. If `None` the maximum range of `x` is
        used. That is only useful if `padding > 0`.
    threshold : float
        The threshold relative to the maximum of `y` of values that should be
        included in the bracket.
    padding : float
        The relative padding on each side in fractions of the bracket size.
    extend : bool, optional
        Signals if the returned range can be larger than the values in ``x``.
        Default is `True`.

    Returns
    -------
    xl, xr : float
        Lowest and biggest value of the range.

    """
    if y is None:
        x1, x2 = np.min(x), np.max(x)
        if not extend:
            return (x1, x2)
    else:
        idx1, idx2 = arglimit(y, threshold=threshold)
        x1, x2 = sorted([x[idx1], x[idx2]])

    # calculate the padding
    if padding != 0.0:
        pad = (x2 - x1) * padding
        x1 -= pad
        x2 += pad

    if not extend:
        x1 = max(x1, np.min(x))
        x2 = min(x2, np.max(x))

    return (x1, x2)


def fwhm(x, y):
    """ Calculates the full width at half maximum of the distribution described
        by (x, y).
    """
    xl, xr = limit(x, y, threshold=0.5, padding=0.0)
    return np.abs(xr - xl)


def edges(x):
    """ Calculates the edges of the array elements.

    Assuming that the input array contains the midpoints of a supposed data
    set, the function returns the (N+1) edges of the data set points.
    """
    diff = np.diff(x)
    reverse = False
    if np.any(np.sign(diff) != np.sign(diff[0])):
        raise ValueError("Input array must be sorted")
    elif diff[0] < 0.0:
        x = x[::-1]
        reverse = True

    result = np.concatenate((
        [1.5 * x[0] - 0.5 * x[1]],
        0.5 * (x[1:] + x[:-1]),
        [1.5 * x[-1] - 0.5 * x[-2]]
    ))
    if reverse:
        result = result[::-1]

    return result


def build_coords(*axes):
    """ Builds a coordinate array from the axes.
    """
    AXES = np.meshgrid(*axes, indexing='ij')
    return np.stack(AXES, axis=-1)


def mask_phase(x, amp, phase, threshold=1e-3):
    mask = (amp / np.max(amp) < threshold)
    blank_phase = np.ma.masked_array(phase, mask=mask)
    blank_x = np.ma.masked_array(x, mask=mask)
    return blank_x, blank_phase

def retrieval_report(res):
    """ Simple helper that prints out important information from the
    retrieval result object.
    """
    print("Retrieval report")
    print("trace error".ljust(15) + "R = %.17e".rjust(25) % res.trace_error)
    if hasattr(res, "trace_error_optimal"):
        print("min. trace error".ljust(15) + "R0 = %.17e".rjust(25) % res.trace_error_optimal)
        print("".ljust(15) + "R - R0 = %.17e".rjust(25) % (res.trace_error - res.trace_error_optimal))
        print()
        print("pulse error".ljust(15) + "ε = %.17e".rjust(25) % res.pulse_error)




#%% Import modules
import femtoQ.tools as fq               # Ez functions
import numpy as np                # Fast numerical operations
import matplotlib.pyplot as plt   # Plotting
from scipy.constants import c as C
from scipy.interpolate import interp1d as interp





def TwoDSI():
    
    #%% Set constants and recurrant functions
    C = 2.998e8                       # Speed of light
    pi = np.pi                        # Pi
    sqrt = lambda x: np.sqrt(x)       # Square root
    log = lambda x: np.log(x)         # Natural logarithm
    
    
    #%% Simulation parameters
    T = 10000e-15             # Time domain range. Higher values wield better frequency domain resolution
    dt = 0.05e-15              # Time increments. Lower values wield better time domain resolution
    
    lambda_1 = 0.5e-6         # Shortest wavelength to include in linear propagation calculations
    lambda_2 = 2e-6           # Longuest wavelength to include in linear propagation calculations
    
    #%% Physics parameters
    tau_FWHM = 8e-15         # Initial pulse duration, full width at half maximum
    lambda0 = 1300e-9         # Carrier wavelength of the pulse
    
    
    chirpingMaterials = ['SF11', 'SF68', 'BK7', 'ZnSe'] # Dispersive elements through which your pulse under study is propagating
    chirpingThicknesses = [ 0.00, 0.0 , 0.000 , 0.0000] # Dispersive materials thicknesses
    
    tdsiMaterial = ['SF11', 'SF68', 'BK7', 'ZnSe']  # Dispersive elements in the long pulse branch of 2DSI
    tdsiThicknesses = [ 0.00, 0.00, 2*2.54e-2+2e-3, 5e-3] # Dispersive materials thicknesses
    
    # Manual values of dispersidon, up to 4th order for now
    GDD = 0 * (1e-15)**2
    TOD = 0 * (1e-15)**3
    FOD =  0* (1e-15)**4
    
    #%% Setup parameters
    
    averageUpconvSpectrum = 1
    
    noiseLevel = 0.0 # Gaussian noise on the simulated spectrometer. Signal peaks at 1
    
    xSpectro = np.linspace(187e-9,1037e-9,2048) # Sampling points of the spectrometer, realistic values could be (187e-9,1037e-9,2048)
    
    ancillaType = "perfect" # Define source to use as ancilla (long pulse / monochromatic-like wave) use "real" to simulate dispersive arm, "perfect" for monochromatic waves
    shearFrequency = 3.5e12 # Shear to set when using "perfect" ancilla
    stageShearScanMinMax = [0e-6, 10e-6] # Min/max position of piezo stage. Values closer to 0 lead to lower shear with "real" ancilla. Width should be at least 3x your center wavelength
    stageTraceScanMinMax = [2e-6, 5e-6] # Min/max position of piezo stage. Values closer to 0 lead to lower shear with "real" ancilla. Width should be at least 3x your center wavelength
    traceNumberDataPoints = 200 # Number of data points to acquire during stage displacement from min to max value
    shearNumberDataPoints = 30 # Number of data points to acquire during stage displacement from min to max value
    
    # %% Quick calculations
    #Pulse parameters
    tau = tau_FWHM / sqrt(2*log(2)) # Gaussian pulse half width @ e^-2
    v0 = C/lambda0                  # Pulse's carrier frequency
    w0 = 2*pi*v0                    # Pulse's carrier angular frequency
    bandwidth = 0.44 / tau_FWHM     # Pulse's FWHM bandwidth
    
    # Time vector
    t = np.linspace(-T/2,T/2, round(T/dt) )
    
    # Time-domain electric field vector
    E = np.exp(1j * w0 * t) * np.exp(-  (t)**2 / (tau)**2)
    
    
    # Frequency domain e-field
    v,s = fq.ezfft(t,E, neg = True)
    
    #%% Dispersion of pulse under study
    
    # Add manual dispersion
    w = 2*pi*v
    s_abs = np.abs(s)
    s_phase = np.unwrap(np.angle(s))
    s_phase +=  (1/2 * GDD * (w-w0-2*pi*5e12)**2) + (1/6 * TOD * (w-w0)**3) + (1/24 * FOD * (w-w0)**4)
    s = s_abs * np.exp(1j*s_phase)
    t, E = fq.ezifft(v,s)
    
    # Add material dispersion
    for ii, material in enumerate(chirpingMaterials):
        L = chirpingThicknesses[ii]
        
        window = LinearWindow(material)
        
        n_sellmeier = window.get_n()
        
        
        
        #%% Fast fourier transform
        
        # v is freqnecy vector, s is frequency-domain electric field vector
        v,s = fq.ezfft(t,E, neg = True)
        
        
        #%% Dispersive propagation
        
        # Convert Sellmeier's equations cutoff wavelengths to frequencies
        v1 = C / lambda_2
        v2 = C / lambda_1
        
        # Convert frequencies within the equation's validity domain to wavelengths in um
        vtmp = v[( (v>v1) & (v<v2) )]
        lambdatmp = C / vtmp
        
        # Calculate refractive index for relevant frequencies
        ntmp = n_sellmeier(lambdatmp*1e6)
        
        # Generate refractive index vector n(v). Values outside of validity range are
        # set to zero.
        n = np.ones_like(v)
        n[( (v>v1) & (v<v2) )] = ntmp
        
        # Apply material's spectral phase to frequency domain electric field
        s[v!=0] = s[v!=0] * np.exp(1j * 2 * pi * n[v!=0] * (v[v!=0] / C) * L)
        
        # Finalpulse's spectral phase
        s_phase = np.unwrap(np.angle(s))
        
        #%% Inverse fast fourier transform
        
        # E2 is final time domain electric field vector, t2 is its corresponding time
        # vector. If all works as intended, t2 is equivalent to t at this point
        t2, E2 = fq.ezifft(v,s)
    
        E = E2
        t = t2   
        
    
    # Find final pulse's peak in time
    tpeak = np.mean( t[abs(E)**2 == np.nanmax(abs(E)**2)] )
    
    tExtended = np.concatenate((t-np.abs(np.min(t)) - np.max(t) - dt, t, t+np.abs(np.min(t)) + np.max(t) + dt), axis = 0 )
    
    EExtended = np.pad(E, E.shape, mode = 'wrap')
    
    
    # Shift time vector to get peak at t=0
    E = np.interp(t, tExtended-tpeak, EExtended)
    
    t2 = t
    E2 = E
    
        
    #%% Simulate 2DSI dispersive arm
    for ii, material in enumerate(tdsiMaterial):
        L = tdsiThicknesses[ii]
        
        window = LinearWindow(material)
        
        n_sellmeier = window.get_n()
        
        
        
        #%% Fast fourier transform
        
        # v is freqnecy vector, s is frequency-domain electric field vector
        v,s = fq.ezfft(t2,E2, neg = True)
        
        
        #%% Dispersive propagation
        
        # Convert Sellmeier's equations cutoff wavelengths to frequencies
        v1 = C / lambda_2
        v2 = C / lambda_1
        
        # Convert frequencies within the equation's validity domain to wavelengths in um
        vtmp = v[( (v>v1) & (v<v2) )]
        lambdatmp = C / vtmp
        
        # Calculate refractive index for relevant frequencies
        ntmp = n_sellmeier(lambdatmp*1e6)
        
        # Generate refractive index vector n(v). Values outside of validity range are
        # set to zero.
        n = np.ones_like(v)
        n[( (v>v1) & (v<v2) )] = ntmp
        
        # Apply material's spectral phase to frequency domain electric field
        s[v!=0] = s[v!=0] * np.exp(1j * 2 * pi * n[v!=0] * (v[v!=0] / C) * L)
        
        # Finalpulse's spectral phase
        s_phase = np.unwrap(np.angle(s))
        
        #%% Inverse fast fourier transform
        
        # E2 is final time domain electric field vector, t2 is its corresponding time
        # vector. If all works as intended, t2 is equivalent to t at this point
        t2, E2 = fq.ezifft(v,s)
        
        
    
    
    
    
    
    #%% Final pulse realignment
    
    # Find final pulse's peak in time
    tpeak = np.mean( t2[abs(E2)**2 == np.nanmax(abs(E2)**2)] )
    
    t2Extended = np.concatenate((t2-np.abs(np.min(t2)) - np.max(t2) - dt, t2, t2+np.abs(np.min(t2)) + np.max(t2) + dt), axis = 0 )
    
    E2Extended = np.pad(E2, E2.shape, mode = 'wrap')
    
    
    # Shift time vector
    E2 = np.interp(t2, t2Extended-tpeak, E2Extended)
    #t2 = t2 - tpeak
        
        
    #%% Initial/Final pulse FWHM duration
      
    #%% Instananeous frequency/wavelength calculations
    
    # Temporal phase
    phi_t = np.unwrap(np.angle(E2))
    
    # Instantaneous angular frequency
    w_inst = (phi_t[2:] - phi_t[:-2]) / (t2[2] - t2[0])
    
    
    #%% Plotting
    
    # Time domain initial/final pulse intensity
    figT = plt.figure()
    axT = figT.gca()
    axT.plot(t*1e15,np.abs(E)**2,'b', label = 'Initial')
    axT.plot(t2*1e15,np.abs(E2)**2,'--r', label = 'Final')
    axT.legend(loc = 'best')
    #axT.set_xlim(-3*max([tau1,tau2])*1e15,3*max([tau1,tau2])*1e15)
    axT.set_xlabel('Time [fs]')
    axT.set_ylabel('Intensity [arb. uni.]')
    
    
    # Spectrum (Intensity + Phase)
    figF = plt.figure()
    axFl = figF.gca()
    axFr = axFl.twinx()
    axFl.plot(v/1e12, np.abs(s)**2, 'b')
    axFr.plot(v/1e12, s_phase, 'r')
    axFl.set_xlim((v0-3*bandwidth)/1e12, (v0 + 3*bandwidth)/1e12)
    axFl.set_xlabel('Frequency [THz]')
    axFl.set_ylabel('Intensity [arb. uni.]', color = 'blue')
    axFr.set_ylim(np.min(s_phase[ ( np.abs(v-v0) < 3*bandwidth)  ]),np.max(s_phase[ ( np.abs(v-v0) < 3*bandwidth)  ]) )
    axFr.set_ylabel('Spectral phase [rad]', color = 'red')
    
    
    # Instantaneous frequency. Time intensity included for comparison
    figT = plt.figure()
    axTl = figT.gca()
    axTr = axTl.twinx()
    axTl.plot(t2*1e15,np.abs(E2)**2,'b')
    axTr.plot(t2[1:-1]*1e15, w_inst/(2*pi)/1e12 ,'r')
    #axTl.set_xlim(-2*tau2*1e15,2*tau2*1e15)
    axTr.set_ylim((v0-2*bandwidth)/1e12, (v0+2*bandwidth)/1e12)
    axTl.set_xlabel('Time [fs]')
    axTl.set_ylabel('Intensity [arb. uni.]', color = 'blue')
    axTr.set_ylabel(r'$\nu_{inst}$ [THz]', color = 'red')
    
    
    tau_sfg = 0e-6 / C 
    tau_cw_vec = np.linspace(stageTraceScanMinMax[0],stageTraceScanMinMax[1],traceNumberDataPoints) / C * 2
    tau_omega = 0e-6 / C 
    
    trace = np.zeros((tau_cw_vec.shape[0], xSpectro.shape[0]))
    
    E = E / np.max(np.sqrt(np.abs(E)**2))
    
    if ancillaType == "real":
        Ecw1 = E2 /  np.max(np.sqrt(np.abs(E2)**2)) 
        Ecw2 = E2 /  np.max(np.sqrt(np.abs(E2)**2))
    elif ancillaType == "perfect":
        Ecw1 = np.exp(1j*w0*t2) 
        Ecw2 = np.exp(1j*(w0+2*np.pi*shearFrequency)*t2)
        
        
    # =============================================================================
    # Ecw1 = E2 /  np.max(np.sqrt(np.abs(E2)**2)) #np.exp(1j*w0*t2) #
    # Ecw2 = E2 /  np.max(np.sqrt(np.abs(E2)**2)) #np.exp(1j*(w0+2*np.pi*2.45e12)*t2) #
    # =============================================================================
    
    for  ii, tau_cw in enumerate(tau_cw_vec):
        
        print(str(ii+1)+'/'+str(tau_cw_vec.shape[0]))
    
        E_short = np.interp(t,t+tau_sfg,E)
        
        E_omega = np.interp(t, t2+tau_omega, Ecw1)
        
        E_cw = np.interp(t, t2+tau_cw, Ecw2)
        
        #Etot = E + E2 + E3
        
        Esq = E_short*E_omega + E_short*E_cw
        
        
        nu, Esqnu = fq.ezfft(t,Esq)
        
        lamb = C/nu
        
        II = np.argsort(lamb)
        lamb = lamb[II]
        Esqlamb = Esqnu[II]
        
        trace[ii,:] = np.interp(xSpectro, lamb, np.abs(Esqlamb)**2) + np.random.normal(scale = noiseLevel, size = xSpectro.shape[0])
        
    trace = trace / np.max(trace)
    
    tau_cw_vec = np.linspace(stageShearScanMinMax[0],stageShearScanMinMax[1],shearNumberDataPoints) / C * 2
    
    
    shearSpectra = np.zeros((tau_cw_vec.shape[0], xSpectro.shape[0]))
    
    for  ii, tau_cw in enumerate(tau_cw_vec):
        
        
        E_short = np.interp(t,t+tau_sfg,E)
        
        E_omega = np.interp(t, t2+tau_omega, Ecw1)
        
        E_cw = np.interp(t, t2+tau_cw, Ecw2)
        
        #Etot = E + E2 + E3
        
        Esq = E_short*E_cw
        
        
        nu, Esqnu = fq.ezfft(t,Esq)
        
        lamb = C/nu
        
        II = np.argsort(lamb)
        lamb = lamb[II]
        Esqlamb = Esqnu[II]
        
        shearSpectra[ii,:] = np.interp(xSpectro, lamb, np.abs(Esqlamb)**2)+ np.random.normal(scale = noiseLevel, size = xSpectro.shape[0])
    shearSpectra = shearSpectra/np.max(shearSpectra)
    
    upconvSpectrum =np.interp(xSpectro,lamb, np.abs( fq.ezfft(t, E_short*E_omega )[1][II] )**2)+ np.random.normal(scale = noiseLevel, size = xSpectro.shape[0])
    
    for ii in range(averageUpconvSpectrum-1):
        upconvSpectrum += np.interp(xSpectro,lamb, np.abs( fq.ezfft(t, E_short*E_omega )[1][II] )**2)+ np.random.normal(scale = noiseLevel, size = xSpectro.shape[0])
    
    upconvSpectrum /= averageUpconvSpectrum
    
    
    upconvSpectrum = upconvSpectrum / np.max(upconvSpectrum)
    
    
    zShear = np.linspace(stageShearScanMinMax[0],stageShearScanMinMax[1],shearNumberDataPoints)*1e6
    zTrace = np.linspace(stageTraceScanMinMax[0],stageTraceScanMinMax[1],traceNumberDataPoints)*1e6
    
    
    plt.figure()
    plt.pcolor(xSpectro*1e9,zTrace,trace)
    
    
    
    plt.figure()
    plt.pcolor(xSpectro*1e9,zShear,shearSpectra)
    
    
    
    
    
    plt.figure()
    plt.plot(xSpectro*1e9,upconvSpectrum)
    
    
    
    
    np.savez('sim2dsiData', wavelengths = xSpectro*1e9, shearStagePosition = zShear, twoDSIStagePosition = zTrace,  upconvSpectrum = upconvSpectrum, shearTrace = shearSpectra, twoDSITrace = trace, timeVector = t, inputPulse = np.abs(E)**2/ np.max(np.abs(E)**2))

    return


def FROG():
    
    pulse = fq.Pulse(800e-9,5e-15,T=1000e-15,dt = 0.05e-15)
    pulse = pulse.disperse(dispVec=[-25,-50,500,2000])
    
    v0 = C/800e-9
    
    v, s = fq.ezfft(pulse.t,pulse.E,neg = True)
    
    s *= 0.2*(v-v0)/v0 -0.3*(v-v0)**2/v0**2
    
    pulse.t,pulse.E = fq.ezifft(v,s)
    
    pulse.E = pulse.E / np.max(np.abs(pulse.E))
    
    
    
    delays = np.linspace(-70e-15,70e-15,512)
    spectroX = np.linspace(300e-9,600e-9,512)
    
    noise_level = 0.1
    
    trace = np.zeros((len(delays),len(spectroX)))
    
    for ii,delay in enumerate(delays):
    
        delayedField = interp(pulse.t+delay,pulse.E,'quadratic',bounds_error=False,fill_value=0)
        pulse2 = fq.Pulse(t=pulse.t,E=delayedField(pulse.t))
        
        vtmp, shgV = fq.ezfft(pulse.t,pulse.E*pulse2.E)
    
        shgV = np.abs(shgV)**2
        interpSHG = interp(C/vtmp[-1::-1],shgV[-1::-1],'quadratic',bounds_error=False,fill_value=0)
        
        trace[ii,:] = interpSHG(spectroX)
        
    trace /= np.max(trace)
    trace += np.random.normal(0,noise_level,trace.shape)
    
    
    plt.figure()
    plt.plot(pulse.t*1e15,np.abs(pulse.E)**2 / np.max(np.abs(pulse.E)**2))
    plt.xlabel('Time [fs]')
    plt.ylabel('Normalized power')
    plt.xlim(-100,100)
    
    v, s = fq.ezfft(pulse.t,pulse.E)
    
    plt.figure()
    plt.plot(C/v*1e9,np.abs(s)**2 / np.max(np.abs(s)**2))
    plt.xlabel('Wavelength [nm]')
    plt.xlim(spectroX[0]*2e9,spectroX[-1]*2e9)
    plt.ylabel('Normalized power density')
    
    plt.figure()
    plt.pcolor(spectroX*1e9,delays*1e15,trace)
    plt.xlabel('Wavelength [nm]')
    plt.ylabel('Delay [fs]')
    c = plt.colorbar()
    c.set_label('Normalized power density')
    
    np.savez('simFROGData.npz',time = delays*1e15, wavelengths = spectroX*1e9, trace = trace)
    
    return




class LinearWindow:
    
    def __init__(self, material):
        
        self.material = material
        
        if   material.lower() == 'sf11':
            self.n = lambda x: (1+1.73759695/(1-0.013188707/x**2)+0.313747346/(1-0.0623068142/x**2)+1.89878101/(1-155.23629/x**2))**.5
        elif material.lower() == 'sf68':
            self.n = lambda x: (1+2.3330067/(1-0.0168838419/x**2)+0.452961396/(1-0.0716086325/x**2)+1.25172339/(1-118.707479/x**2))**.5
        elif material.lower() == 'bk7':
            self.n = lambda x: (1+1.03961212/(1-0.00600069867/x**2)+0.231792344/(1-0.0200179144/x**2)+1.01046945/(1-103.560653/x**2))**.5
        elif material.lower() == 'znse':
            self.n = lambda x: (1+3.00+1.90/(1-0.113/x**2))**.5
            
    def get_n(self):
        return self.n


import numpy as np
from . import lib


def autocorrelation(pulse, tau=None, collinear=False):
    ''' Calculates the intensity or second-order autocorrelation G2(tau).

    Parameters
    ----------
    pulse : Pulse instance
        The pulse of which the autocorrelation is calculated
    tau : 1d-array, optional
        The delays at which the autocorrelation is evaluated. If `None`
        the temporal grid `pulse.t` of the pulse object is used.
    collinear : bool, optional
        Calculates the collinear autocorrelation, with background and higher
        frequency terms. Otherwise only the non-collinear intensity
        autocorrelation without background is calculated. Default is `False`.

    Returns
    -------
    tau : 1d-array
        The delay axis at which the autocorrelation was evaluated.
    ac : 1d-array
        The autocorrelation signal.

    Notes
    -----
    Calculates the following expression::

        G2(tau) = int | [E(t-tau) + E(t)]^2 |^2 dt

    where E(t) is the real-valued electric field.
    This expression is expanded in terms of the complex-valued pulse envelope
    and then evaluated with help of the convolution theorem and the Fourier
    transform. Terms containing oscillations in t, e.g., exp(2j w0 t), are
    neglected.
    Specifically, it implements Eq. (9.7) on page 460 from [Diels2006]_.
    '''
    ft, ift = pulse.ft.forward, pulse.ft.backward
    ift_at = pulse.ft.backward_at

    # if the delays are not provided, use the pulse grid
    fft = False
    if tau is None:
        fft = True  # in this case we can use the FFT directly
        tau = pulse.t

    if collinear:
        f = pulse.field
        i = lib.abs2(f)  # intensity
        if fft:
            # uses the n log(n) FFT
            a0 = ift(lib.abs2(ft(pulse.intensity)))
            a1 = ift(np.real(ft(f).conj() * ft(f * i)))
            a2 = ift(lib.abs2(ft(f * f)))
        else:
            # uses the n^2 DFT
            a0 = ift_at(lib.abs2(ft(pulse.intensity)), tau)
            a1 = ift_at(np.real(ft(f).conj() * ft(f * i)), tau)
            a2 = ift_at(lib.abs2(ft(f * f)), tau)

        ac = (  4.0 * np.real(a0)  # A0
              + pulse.dt * (i * i).sum() / np.pi  # background
              + 8.0 * np.real(a1 * np.exp(1.0j * pulse.w0 * tau))  # A1
              + 2.0 * np.real(a2 * np.exp(2.0j * pulse.w0 * tau)) )  # A2
        # scale for classic 8:1 ratio
        ac *= 8.0 / np.max(ac)
    else:
        i = pulse.intensity
        ac_omega = 4.0 * lib.abs2(ft(i))
        if fft:
            ac = ift(ac_omega).real
        else:
            ac = ift_at(ac_omega, tau).real
        # normalize
        ac /= np.max(ac)

    return tau, ac
